import datetime
import logging
import os
import subprocess
import uuid

import mock
import pytest

from uaclient import apt, exceptions, messages, system


class TestGetKernelInfo:
    @pytest.mark.parametrize(
        [
            "uname_machine",
            "uname_release",
            "proc_version_signature_side_effect",
            "build_date",
            "expected",
        ],
        (
            (
                "x86_64",
                "5.14.0-1024-oem",
                "Ubuntu 5.14.0-1024.26-oem 5.15.100",
                datetime.datetime(2023, 4, 6, 7, 48, 48),
                system.KernelInfo(
                    uname_machine_arch="x86_64",
                    uname_release="5.14.0-1024-oem",
                    proc_version_signature_version="5.14.0-1024.26-oem",
                    build_date=datetime.datetime(2023, 4, 6, 7, 48, 48),
                    major=5,
                    minor=14,
                    patch=0,
                    abi="1024",
                    flavor="oem",
                ),
            ),
            (
                "aarch64",
                "5.14.0-1024-oem",
                "Ubuntu 5.14.0-1024.26-oem 5.15.100",
                datetime.datetime(2023, 4, 6, 7, 48, 48),
                system.KernelInfo(
                    uname_machine_arch="aarch64",
                    uname_release="5.14.0-1024-oem",
                    proc_version_signature_version="5.14.0-1024.26-oem",
                    build_date=datetime.datetime(2023, 4, 6, 7, 48, 48),
                    major=5,
                    minor=14,
                    patch=0,
                    abi="1024",
                    flavor="oem",
                ),
            ),
            (
                "x86_64",
                "4.4.0-21-generic",
                "Ubuntu 4.4.0-21.37-generic 4.15.100",
                datetime.datetime(2023, 4, 6, 7, 48, 48),
                system.KernelInfo(
                    uname_machine_arch="x86_64",
                    uname_release="4.4.0-21-generic",
                    proc_version_signature_version="4.4.0-21.37-generic",
                    build_date=datetime.datetime(2023, 4, 6, 7, 48, 48),
                    major=4,
                    minor=4,
                    patch=0,
                    abi="21",
                    flavor="generic",
                ),
            ),
            (
                "x86_64",
                "5.4.0-52-generic",
                "Ubuntu 5.4.0-52.37-generic 5.15.100",
                datetime.datetime(2023, 4, 6, 7, 48, 48),
                system.KernelInfo(
                    uname_machine_arch="x86_64",
                    uname_release="5.4.0-52-generic",
                    proc_version_signature_version="5.4.0-52.37-generic",
                    build_date=datetime.datetime(2023, 4, 6, 7, 48, 48),
                    major=5,
                    minor=4,
                    patch=0,
                    abi="52",
                    flavor="generic",
                ),
            ),
            (
                "x86_64",
                "5.4.0-52-generic",
                "Ubuntu 5.4.0-52.37~20.04-generic 5.15.100",
                datetime.datetime(2023, 4, 6, 7, 48, 48),
                system.KernelInfo(
                    uname_machine_arch="x86_64",
                    uname_release="5.4.0-52-generic",
                    proc_version_signature_version="5.4.0-52.37~20.04-generic",
                    build_date=datetime.datetime(2023, 4, 6, 7, 48, 48),
                    major=5,
                    minor=4,
                    patch=0,
                    abi="52",
                    flavor="generic",
                ),
            ),
            (
                "x86_64",
                "5.4.0-52-generic",
                Exception(),
                datetime.datetime(2023, 4, 6, 7, 48, 48),
                system.KernelInfo(
                    uname_machine_arch="x86_64",
                    uname_release="5.4.0-52-generic",
                    proc_version_signature_version=None,
                    build_date=datetime.datetime(2023, 4, 6, 7, 48, 48),
                    major=5,
                    minor=4,
                    patch=0,
                    abi="52",
                    flavor="generic",
                ),
            ),
            (
                "x86_64",
                "5.4.0-1021-aws-fips",
                "Ubuntu 5.4.0-1021.21+fips2-aws-fips 5.4.44",
                datetime.datetime(2023, 4, 6, 7, 48, 48),
                system.KernelInfo(
                    uname_machine_arch="x86_64",
                    uname_release="5.4.0-1021-aws-fips",
                    proc_version_signature_version="5.4.0-1021.21+fips2-aws-fips",  # noqa: E501
                    build_date=datetime.datetime(2023, 4, 6, 7, 48, 48),
                    major=5,
                    minor=4,
                    patch=0,
                    abi="1021",
                    flavor="aws-fips",
                ),
            ),
            (
                "x86_64",
                "4.4.0-1017-fips",
                "Ubuntu 4.4.0-1017.22~recert1-fips 4.4.185",
                datetime.datetime(2023, 4, 6, 7, 48, 48),
                system.KernelInfo(
                    uname_machine_arch="x86_64",
                    uname_release="4.4.0-1017-fips",
                    proc_version_signature_version="4.4.0-1017.22~recert1-fips",  # noqa: E501
                    build_date=datetime.datetime(2023, 4, 6, 7, 48, 48),
                    major=4,
                    minor=4,
                    patch=0,
                    abi="1017",
                    flavor="fips",
                ),
            ),
            (
                "x86_64",
                "4.4.0-1017.something.invalid-fips",
                "Ubuntu 4.4.0-1017.22~recert1-fips 4.4.185",
                datetime.datetime(2023, 4, 6, 7, 48, 48),
                system.KernelInfo(
                    uname_machine_arch="x86_64",
                    uname_release="4.4.0-1017.something.invalid-fips",
                    proc_version_signature_version="4.4.0-1017.22~recert1-fips",  # noqa: E501
                    build_date=datetime.datetime(2023, 4, 6, 7, 48, 48),
                    major=None,
                    minor=None,
                    patch=None,
                    abi=None,
                    flavor=None,
                ),
            ),
            (
                "x86_64",
                "4.4.0-1017.something.invalid-fips",
                "Ubuntu 4.4.0-1017.22~recert1-fips 4.4.185",
                None,
                system.KernelInfo(
                    uname_machine_arch="x86_64",
                    uname_release="4.4.0-1017.something.invalid-fips",
                    proc_version_signature_version="4.4.0-1017.22~recert1-fips",  # noqa: E501
                    build_date=None,
                    major=None,
                    minor=None,
                    patch=None,
                    abi=None,
                    flavor=None,
                ),
            ),
        ),
    )
    @mock.patch("uaclient.system._get_kernel_build_date")
    @mock.patch("uaclient.system.load_file")
    @mock.patch("uaclient.system.os.uname")
    def test_get_kernel_info(
        self,
        m_uname,
        m_load_file,
        m_get_kernel_build_date,
        uname_machine,
        uname_release,
        proc_version_signature_side_effect,
        build_date,
        expected,
    ):
        m_uname.return_value = mock.MagicMock(
            release=uname_release, machine=uname_machine
        )
        m_load_file.side_effect = [proc_version_signature_side_effect]
        m_get_kernel_build_date.return_value = build_date
        assert system.get_kernel_info.__wrapped__() == expected

    @pytest.mark.parametrize(
        [
            "uname_result",
            "changelog_timestamp",
            "expected",
        ],
        [
            (
                os.uname_result(
                    [
                        "",
                        "",
                        "",
                        "#20-Ubuntu SMP PREEMPT_DYNAMIC Thu Apr  6 07:48:48 UTC 2023",  # noqa: E501
                        "",
                    ]
                ),
                mock.sentinel.changelog_timestamp,
                datetime.datetime(
                    2023, 4, 6, 7, 48, 48, tzinfo=datetime.timezone.utc
                ),
            ),
            (
                os.uname_result(
                    [
                        "",
                        "",
                        "",
                        "#33~22.04.1-Ubuntu SMP PREEMPT_DYNAMIC Mon Jan 30 17:03:34 UTC 2",  # noqa: E501
                        "",
                    ]
                ),
                mock.sentinel.changelog_timestamp,
                mock.sentinel.changelog_timestamp,
            ),
            (
                os.uname_result(["", "", "", "corrupted", ""]),
                mock.sentinel.changelog_timestamp,
                mock.sentinel.changelog_timestamp,
            ),
        ],
    )
    @mock.patch("uaclient.system._get_kernel_changelog_timestamp")
    def test_get_kernel_build_date(
        self,
        m_get_kernel_changelog_timestamp,
        uname_result,
        changelog_timestamp,
        expected,
    ):
        m_get_kernel_changelog_timestamp.return_value = changelog_timestamp
        assert expected == system._get_kernel_build_date(uname_result)

    @pytest.mark.parametrize(
        [
            "uname_result",
            "is_container",
            "stat_result",
            "expected_stat_call_args",
            "expected",
        ],
        [
            (
                None,
                True,
                None,
                [],
                None,
            ),
            (
                os.uname_result(["", "", "version-here", "", ""]),
                False,
                Exception(),
                [
                    mock.call(
                        "/usr/share/doc/linux-image-version-here/changelog.Debian.gz"  # noqa: E501
                    )
                ],
                None,
            ),
            (
                os.uname_result(["", "", "version-here", "", ""]),
                False,
                [os.stat_result([0, 0, 0, 0, 0, 0, 0, 0, 1680762951, 0])],
                [
                    mock.call(
                        "/usr/share/doc/linux-image-version-here/changelog.Debian.gz"  # noqa: E501
                    )
                ],
                datetime.datetime(
                    2023, 4, 6, 6, 35, 51, tzinfo=datetime.timezone.utc
                ),
            ),
        ],
    )
    @mock.patch("os.stat")
    @mock.patch("uaclient.system.is_container")
    def test_get_kernel_changelog_timestamp(
        self,
        m_is_container,
        m_os_stat,
        uname_result,
        is_container,
        stat_result,
        expected_stat_call_args,
        expected,
    ):
        m_is_container.return_value = is_container
        m_os_stat.side_effect = stat_result
        assert expected == system._get_kernel_changelog_timestamp(uname_result)
        assert expected_stat_call_args == m_os_stat.call_args_list


class TestGetDpkgArch:
    @pytest.mark.parametrize(
        "stdout, expected",
        (
            (
                "amd64",
                "amd64",
            ),
            (
                "arm64\n",
                "arm64",
            ),
            (
                "   arm64    \n",
                "arm64",
            ),
        ),
    )
    @mock.patch("uaclient.system.subp")
    def test_get_dpkg_arch(self, m_subp, stdout, expected):
        m_subp.return_value = (stdout, "")
        assert system.get_dpkg_arch.__wrapped__() == expected
        assert m_subp.call_args_list == [
            mock.call(["dpkg", "--print-architecture"])
        ]


class TestGetVirtType:
    @pytest.mark.parametrize(
        [
            "subp_side_effect",
            "load_file_side_effect",
            "expected",
        ],
        [
            ([("", "")], None, ""),
            ([("lxc\n", "")], None, "lxc"),
            ([("lxc\n", "anything")], None, "lxc"),
            (
                exceptions.ProcessExecutionError(cmd="", stdout="", stderr=""),
                [""],
                "",
            ),
            (
                exceptions.ProcessExecutionError(cmd="", stdout="", stderr=""),
                ["line\nline\nlinedockerline\nline"],
                "docker",
            ),
            (
                exceptions.ProcessExecutionError(cmd="", stdout="", stderr=""),
                ["line\nline\nlinebuildkitline\nline"],
                "docker",
            ),
            (
                exceptions.ProcessExecutionError(cmd="", stdout="", stderr=""),
                ["line\nline\nlinebuildahline\nline"],
                "podman",
            ),
            (
                exceptions.ProcessExecutionError(cmd="", stdout="", stderr=""),
                ["line\nline\nline\nline"],
                "",
            ),
            (
                exceptions.ProcessExecutionError(cmd="", stdout="", stderr=""),
                FileNotFoundError(),
                "",
            ),
        ],
    )
    @mock.patch("uaclient.system.load_file")
    @mock.patch("uaclient.system.subp")
    def test_get_virt_type(
        self,
        m_subp,
        m_load_file,
        subp_side_effect,
        load_file_side_effect,
        expected,
    ):
        m_subp.side_effect = subp_side_effect
        m_load_file.side_effect = load_file_side_effect
        assert expected == system.get_virt_type.__wrapped__()


class TestIsLTS:
    @pytest.mark.parametrize(
        "series, supported_esm, expected",
        (
            ("xenial", "xenial\nbionic\nfocal", True),
            ("groovy", "xenial\nbionic\nfocal", False),
        ),
    )
    @mock.patch("uaclient.system.subp")
    def test_is_lts_if_distro_info_supported_esm(
        self, subp, series, supported_esm, expected
    ):
        subp.return_value = supported_esm, ""
        # Use __wrapped__ to avoid hitting the lru_cached value across tests
        assert expected is system.is_lts.__wrapped__(series)
        assert [
            mock.call(["/usr/bin/ubuntu-distro-info", "--supported-esm"])
        ] == subp.call_args_list


class TestIsSupported:
    @pytest.mark.parametrize(
        "series,expected", (("sup1", True), ("unsup", False))
    )
    @mock.patch("uaclient.system.subp")
    def test_return_supported_series(self, subp, series, expected):
        subp.return_value = "sup1\nsup2\nsup3", ""
        assert expected is system.is_supported.__wrapped__(series)
        assert [
            mock.call(["/usr/bin/ubuntu-distro-info", "--supported"])
        ] == subp.call_args_list


class TestIsActiveESM:
    @pytest.mark.parametrize(
        "series, is_lts, days_until_esm,expected",
        (
            ("xenial", True, 1, False),
            ("xenial", True, 0, True),
            ("bionic", True, 1, False),
            ("bionic", True, 0, True),
            ("focal", True, 1, False),
            ("focal", True, 0, True),
            ("groovy", False, 0, False),
        ),
    )
    @mock.patch("uaclient.system.subp")
    @mock.patch("uaclient.system.is_lts")
    def test_true_when_supported_esm_release_and_active(
        self, util_is_lts, subp, series, is_lts, days_until_esm, expected
    ):
        """Return True when series is --suported-esm and days until esm < 1."""
        util_is_lts.return_value = is_lts
        subp.return_value = str(days_until_esm), ""

        # Use __wrapped__ to avoid hitting the lru_cached value across tests
        calls = []
        if is_lts:
            calls.append(
                mock.call(
                    [
                        "/usr/bin/ubuntu-distro-info",
                        "--series",
                        series,
                        "-yeol",
                    ]
                )
            )
        assert expected is system.is_active_esm.__wrapped__(series)
        assert calls == subp.call_args_list


class TestIsContainer:
    @mock.patch("uaclient.system.subp")
    def test_true_systemd_detect_virt_success(self, m_subp):
        """Return True when systemd-detect virt exits success."""
        system.is_container.cache_clear()
        m_subp.side_effect = [
            exceptions.ProcessExecutionError(
                "Failed running command 'ischroot' [exit(1)]"
            ),
            "",
            "",
        ]
        assert True is system.is_container()
        # Second call for lru_cache test
        system.is_container()
        calls = [
            mock.call(["ischroot"]),
            mock.call(["systemd-detect-virt", "--quiet", "--container"]),
        ]
        assert calls == m_subp.call_args_list

    @mock.patch("uaclient.system.subp")
    def test_true_on_run_container_type(self, m_subp, tmpdir):
        """Return True when /run/container_type exists."""
        system.is_container.cache_clear()
        m_subp.side_effect = [
            exceptions.ProcessExecutionError(
                "Failed running command 'ischroot' [exit(1)]"
            ),
            OSError("No systemd-detect-virt utility"),
        ]
        tmpdir.join("container_type").write("")

        assert True is system.is_container(run_path=tmpdir.strpath)
        calls = [
            mock.call(["ischroot"]),
            mock.call(["systemd-detect-virt", "--quiet", "--container"]),
        ]
        assert calls == m_subp.call_args_list

    @mock.patch("uaclient.system.subp")
    def test_true_on_run_systemd_container(self, m_subp, tmpdir):
        """Return True when /run/systemd/container exists."""
        system.is_container.cache_clear()
        m_subp.side_effect = [
            exceptions.ProcessExecutionError(
                "Failed running command 'ischroot' [exit(1)]"
            ),
            OSError("No systemd-detect-virt utility"),
        ]
        tmpdir.join("systemd/container").write("", ensure=True)

        assert True is system.is_container(run_path=tmpdir.strpath)
        calls = [
            mock.call(["ischroot"]),
            mock.call(["systemd-detect-virt", "--quiet", "--container"]),
        ]
        assert calls == m_subp.call_args_list

    @mock.patch("uaclient.system.subp")
    def test_false_on_non_sytemd_detect_virt_and_no_runfiles(
        self, m_subp, tmpdir
    ):
        """Return False when sytemd-detect-virt erros and no /run/* files."""
        system.is_container.cache_clear()
        m_subp.side_effect = [
            exceptions.ProcessExecutionError(
                "Failed running command 'ischroot' [exit(1)]"
            ),
            OSError("No systemd-detect-virt utility"),
        ]
        tmpdir.join("systemd/container").write("", ensure=True)

        with mock.patch("uaclient.util.os.path.exists") as m_exists:
            m_exists.return_value = False
            assert False is system.is_container(run_path=tmpdir.strpath)
        calls = [
            mock.call(["ischroot"]),
            mock.call(["systemd-detect-virt", "--quiet", "--container"]),
        ]
        assert calls == m_subp.call_args_list
        exists_calls = [
            mock.call(tmpdir.join("container_type").strpath),
            mock.call(tmpdir.join("systemd/container").strpath),
        ]
        assert exists_calls == m_exists.call_args_list

    @mock.patch("uaclient.system.subp")
    def test_false_on_chroot_system(self, m_subp):
        system.is_container.cache_clear()
        m_subp.return_value = ("", "")
        assert False is system.is_container()

        calls = [mock.call(["ischroot"])]
        assert calls == m_subp.call_args_list


class TestIsDesktop:
    @pytest.mark.parametrize(
        ["installed_packages", "expected"],
        [
            ([], False),
            (
                [
                    apt.InstalledAptPackage(
                        name="not-desktop", version="", arch=""
                    ),
                    apt.InstalledAptPackage(
                        name="desktop-not-ubuntu", version="", arch=""
                    ),
                ],
                False,
            ),
            (
                [
                    apt.InstalledAptPackage(
                        name="not-desktop", version="", arch=""
                    ),
                    apt.InstalledAptPackage(
                        name="ubuntu-desktop", version="", arch=""
                    ),
                ],
                True,
            ),
            (
                [
                    apt.InstalledAptPackage(
                        name="not-desktop", version="", arch=""
                    ),
                    apt.InstalledAptPackage(
                        name="ubuntu-desktop-minimal", version="", arch=""
                    ),
                ],
                True,
            ),
            (
                [
                    apt.InstalledAptPackage(
                        name="not-desktop", version="", arch=""
                    ),
                    apt.InstalledAptPackage(
                        name="kubuntu-desktop", version="", arch=""
                    ),
                ],
                True,
            ),
            (
                [
                    apt.InstalledAptPackage(
                        name="not-desktop", version="", arch=""
                    ),
                    apt.InstalledAptPackage(
                        name="xubuntu-desktop", version="", arch=""
                    ),
                ],
                True,
            ),
            (
                [
                    apt.InstalledAptPackage(
                        name="not-desktop", version="", arch=""
                    ),
                    apt.InstalledAptPackage(
                        name="lubuntu-desktop", version="", arch=""
                    ),
                ],
                True,
            ),
        ],
    )
    @mock.patch("uaclient.apt.get_installed_packages")
    def test_true_systemd_detect_virt_success(
        self,
        m_installed_packages,
        installed_packages,
        expected,
    ):
        m_installed_packages.return_value = installed_packages
        assert expected == system.is_desktop.__wrapped__()


class TestParseOSRelease:
    @pytest.mark.parametrize(
        "content, expected",
        (
            (
                """\
NAME="Ubuntu"
VERSION="16.04.5 LTS (Xenial Xerus)"
ID=ubuntu
ID_LIKE=debian
PRETTY_NAME="Ubuntu 16.04.5 LTS"
VERSION_ID="16.04"
HOME_URL="http://www.ubuntu.com/"
SUPPORT_URL="http://help.ubuntu.com/"
BUG_REPORT_URL="http://bugs.launchpad.net/ubuntu/"
VERSION_CODENAME=xenial
UBUNTU_CODENAME=xenial
""",
                {
                    "NAME": "Ubuntu",
                    "VERSION": "16.04.5 LTS (Xenial Xerus)",
                    "ID": "ubuntu",
                    "ID_LIKE": "debian",
                    "PRETTY_NAME": "Ubuntu 16.04.5 LTS",
                    "VERSION_ID": "16.04",
                    "HOME_URL": "http://www.ubuntu.com/",
                    "SUPPORT_URL": "http://help.ubuntu.com/",
                    "BUG_REPORT_URL": "http://bugs.launchpad.net/ubuntu/",
                    "VERSION_CODENAME": "xenial",
                    "UBUNTU_CODENAME": "xenial",
                },
            ),
            (
                """\
NAME="Ubuntu"
VERSION="18.04.1 LTS (Bionic Beaver)"
ID=ubuntu
ID_LIKE=debian
PRETTY_NAME="Ubuntu 18.04.1 LTS"
VERSION_ID="18.04"
HOME_URL="https://www.ubuntu.com/"
SUPPORT_URL="https://help.ubuntu.com/"
BUG_REPORT_URL="https://bugs.launchpad.net/ubuntu/"
PRIVACY_POLICY_URL="https://www.ubuntu.com/legal/terms-and-policies/privacy-policy"
VERSION_CODENAME=bionic
UBUNTU_CODENAME=bionic
""",
                {
                    "NAME": "Ubuntu",
                    "VERSION": "18.04.1 LTS (Bionic Beaver)",
                    "ID": "ubuntu",
                    "ID_LIKE": "debian",
                    "PRETTY_NAME": "Ubuntu 18.04.1 LTS",
                    "VERSION_ID": "18.04",
                    "HOME_URL": "https://www.ubuntu.com/",
                    "SUPPORT_URL": "https://help.ubuntu.com/",
                    "BUG_REPORT_URL": "https://bugs.launchpad.net/ubuntu/",
                    "PRIVACY_POLICY_URL": "https://www.ubuntu.com/legal/terms-and-policies/privacy-policy",  # noqa: E501
                    "VERSION_CODENAME": "bionic",
                    "UBUNTU_CODENAME": "bionic",
                },
            ),
            (
                """\
PRETTY_NAME="Ubuntu 22.04.1 LTS"
NAME="Ubuntu"
VERSION_ID="22.04"
VERSION="22.04.1 LTS (Jammy Jellyfish)"
VERSION_CODENAME=jammy
ID=ubuntu
ID_LIKE=debian
HOME_URL="https://www.ubuntu.com/"
SUPPORT_URL="https://help.ubuntu.com/"
BUG_REPORT_URL="https://bugs.launchpad.net/ubuntu/"
PRIVACY_POLICY_URL="https://www.ubuntu.com/legal/terms-and-policies/privacy-policy"
UBUNTU_CODENAME=jammy
""",
                {
                    "PRETTY_NAME": "Ubuntu 22.04.1 LTS",
                    "NAME": "Ubuntu",
                    "VERSION_ID": "22.04",
                    "VERSION": "22.04.1 LTS (Jammy Jellyfish)",
                    "VERSION_CODENAME": "jammy",
                    "ID": "ubuntu",
                    "ID_LIKE": "debian",
                    "HOME_URL": "https://www.ubuntu.com/",
                    "SUPPORT_URL": "https://help.ubuntu.com/",
                    "BUG_REPORT_URL": "https://bugs.launchpad.net/ubuntu/",
                    "PRIVACY_POLICY_URL": "https://www.ubuntu.com/legal/terms-and-policies/privacy-policy",  # noqa: E501
                    "UBUNTU_CODENAME": "jammy",
                },
            ),
            (
                """\
PRETTY_NAME="Ubuntu Kinetic Kudu (development branch)"
NAME="Ubuntu"
VERSION_ID="22.10"
VERSION="22.10 (Kinetic Kudu)"
VERSION_CODENAME=kinetic
ID=ubuntu
ID_LIKE=debian
HOME_URL="https://www.ubuntu.com/"
SUPPORT_URL="https://help.ubuntu.com/"
BUG_REPORT_URL="https://bugs.launchpad.net/ubuntu/"
PRIVACY_POLICY_URL="https://www.ubuntu.com/legal/terms-and-policies/privacy-policy"
UBUNTU_CODENAME=kinetic
LOGO=ubuntu-logo
""",
                {
                    "PRETTY_NAME": "Ubuntu Kinetic Kudu (development branch)",
                    "NAME": "Ubuntu",
                    "VERSION_ID": "22.10",
                    "VERSION": "22.10 (Kinetic Kudu)",
                    "VERSION_CODENAME": "kinetic",
                    "ID": "ubuntu",
                    "ID_LIKE": "debian",
                    "HOME_URL": "https://www.ubuntu.com/",
                    "SUPPORT_URL": "https://help.ubuntu.com/",
                    "BUG_REPORT_URL": "https://bugs.launchpad.net/ubuntu/",
                    "PRIVACY_POLICY_URL": "https://www.ubuntu.com/legal/terms-and-policies/privacy-policy",  # noqa: E501
                    "UBUNTU_CODENAME": "kinetic",
                    "LOGO": "ubuntu-logo",
                },
            ),
        ),
    )
    @mock.patch("uaclient.system.load_file")
    def test_parse_os_release(self, m_load_file, content, expected):
        """_parse_os_release returns a dict of values from /etc/os-release."""
        m_load_file.return_value = content
        assert expected == system._parse_os_release.__wrapped__()
        assert m_load_file.call_args_list == [mock.call("/etc/os-release")]


@mock.patch("uaclient.system.load_file")
class TestGetDistroInfo:
    content = """\
version,codename,series,created,release,eol,eol-server,eol-esm
16.04 LTS,Xenial Xerus,xenial,2015-10-22,2016-04-21,2021-04-21,2021-04-21,2024-04-23
20.04 LTS,Focal Fossa,focal,2019-10-17,2020-04-23,2025-04-23,2025-04-23,2030-04-23
22.10,Kinetic Kudu,kinetic,2022-04-21,2022-10-20,2023-07-20
"""  # noqa: E501

    def test_get_distro_info(self, m_load_file):
        m_load_file.return_value = self.content

        xenial_di = system.get_distro_info.__wrapped__("xenial")
        assert 4 == xenial_di.eol.month
        assert 2021 == xenial_di.eol.year
        assert 4 == xenial_di.eol_esm.month
        assert 2026 == xenial_di.eol_esm.year

        focal_di = system.get_distro_info.__wrapped__("focal")
        assert 4 == focal_di.eol.month
        assert 2025 == focal_di.eol.year
        assert 4 == focal_di.eol_esm.month
        assert 2030 == focal_di.eol_esm.year

        kinetic_di = system.get_distro_info.__wrapped__("kinetic")
        assert 7 == kinetic_di.eol.month
        assert 2023 == kinetic_di.eol.year
        assert 7 == kinetic_di.eol_esm.month
        assert 2023 == kinetic_di.eol_esm.year

        with pytest.raises(exceptions.UbuntuProError) as excinfo:
            system.get_distro_info.__wrapped__("nonexistent")
        assert (
            messages.E_MISSING_SERIES_IN_DISTRO_INFO_FILE.format(
                series="nonexistent"
            )
            == excinfo.value.named_msg
        )

    def test_no_csv_file(self, m_load_file):
        m_load_file.side_effect = FileNotFoundError

        with pytest.raises(exceptions.UbuntuProError) as excinfo:
            system.get_distro_info.__wrapped__("focal")
        assert messages.E_MISSING_DISTRO_INFO_FILE == excinfo.value.named_msg


class TestGetReleaseInfo:
    @pytest.mark.parametrize(
        ["version", "expected_exception"],
        (
            (
                "junk",
                exceptions.ParsingErrorOnOSReleaseFile,
            ),
            (
                "22.04 LTS",
                exceptions.MissingSeriesOnOSReleaseFile,
            ),
        ),
    )
    @mock.patch("uaclient.system._parse_os_release")
    def test_get_release_info_error(
        self, m_parse_os_release, version, expected_exception
    ):
        """get_release_info errors when it cannot parse os-release."""
        m_parse_os_release.return_value = {"VERSION": version}
        with pytest.raises(expected_exception):
            # Use __wrapped__ to avoid hitting the
            # lru_cached value across tests
            system.get_release_info.__wrapped__()

    @pytest.mark.parametrize(
        ["os_release", "expected"],
        [
            (
                {
                    "NAME": "Ubuntu",
                    "VERSION": "16.04.5 LTS (Xenial Xerus)",
                },
                system.ReleaseInfo(
                    distribution="Ubuntu",
                    release="16.04",
                    series="xenial",
                    pretty_version="16.04 LTS (Xenial Xerus)",
                ),
            ),
            (
                {
                    "NAME": "Ubuntu",
                    "VERSION": "18.04.1 LTS (Bionic Beaver)",
                },
                system.ReleaseInfo(
                    distribution="Ubuntu",
                    release="18.04",
                    series="bionic",
                    pretty_version="18.04 LTS (Bionic Beaver)",
                ),
            ),
            (
                {
                    "NAME": "Ubuntu",
                    "VERSION": "22.04.1 LTS (Jammy Jellyfish)",
                },
                system.ReleaseInfo(
                    distribution="Ubuntu",
                    release="22.04",
                    series="jammy",
                    pretty_version="22.04 LTS (Jammy Jellyfish)",
                ),
            ),
            (
                {
                    "NAME": "Ubuntu",
                    "VERSION": "22.10 (Kinetic Kudu)",
                },
                system.ReleaseInfo(
                    distribution="Ubuntu",
                    release="22.10",
                    series="kinetic",
                    pretty_version="22.10 (Kinetic Kudu)",
                ),
            ),
            (
                {
                    "NAME": "Ubuntu",
                    "VERSION": "22.04 LTS",
                    "VERSION_CODENAME": "Jammy",
                },
                system.ReleaseInfo(
                    distribution="Ubuntu",
                    release="22.04",
                    series="jammy",
                    pretty_version="22.04 LTS",
                ),
            ),
            (
                {
                    "NAME": "Ubuntu",
                    "VERSION": "CORRUPTED",
                    "VERSION_CODENAME": "Jammy",
                    "VERSION_ID": "22.04",
                },
                system.ReleaseInfo(
                    distribution="Ubuntu",
                    release="22.04",
                    series="jammy",
                    pretty_version="CORRUPTED",
                ),
            ),
        ],
    )
    @mock.patch("uaclient.system.get_kernel_info")
    @mock.patch("uaclient.system.get_dpkg_arch")
    @mock.patch("uaclient.system._parse_os_release")
    @mock.patch("uaclient.system.get_virt_type")
    def test_get_release_info_with_version(
        self,
        m_get_virt_type,
        m_parse_os_release,
        m_get_dpkg_arch,
        m_get_kernel_info,
        os_release,
        expected,
    ):
        m_parse_os_release.return_value = os_release
        assert expected == system.get_release_info.__wrapped__()


@mock.patch("uaclient.files.state_files.machine_id_file.write")
class TestGetMachineId:
    def test_get_machine_id_from_machine_token_file(
        self, _m_machine_id_file_write, fake_machine_token_file
    ):
        fake_machine_token_file.attached = True
        value = system.get_machine_id(None)
        assert "test_machine_id" == value

    @mock.patch("uaclient.files.state_files.machine_id_file.read")
    def test_get_machine_id_from_etc_machine_id(
        self,
        m_machine_id_file_read,
        _m_machine_id_file_write,
        FakeConfig,
        tmpdir,
    ):
        """Presence of /etc/machine-id is returned if it exists."""
        etc_machine_id = tmpdir.join("etc-machine-id")
        m_machine_id_file_read.return_value = None
        assert "/etc/machine-id" == system.ETC_MACHINE_ID
        etc_machine_id.write("etc-machine-id")
        cfg = FakeConfig()
        with mock.patch(
            "uaclient.system.ETC_MACHINE_ID", etc_machine_id.strpath
        ):
            value = system.get_machine_id(cfg)
            # Test lru_cache caches /etc/machine-id from first read
            etc_machine_id.write("does-not-change")
            cached_value = system.get_machine_id(cfg)
            assert value == cached_value
        assert "etc-machine-id" == value

    @mock.patch("uaclient.files.state_files.machine_id_file.read")
    def test_get_machine_id_from_var_lib_dbus_machine_id(
        self,
        m_machine_id_file_read,
        _m_machine_id_file_write,
        FakeConfig,
        tmpdir,
    ):
        """fallback to /var/lib/dbus/machine-id"""
        m_machine_id_file_read.return_value = None
        etc_machine_id = tmpdir.join("etc-machine-id")
        dbus_machine_id = tmpdir.join("dbus-machine-id")
        assert "/var/lib/dbus/machine-id" == system.DBUS_MACHINE_ID
        dbus_machine_id.write("dbus-machine-id")
        cfg = FakeConfig()
        with mock.patch(
            "uaclient.system.DBUS_MACHINE_ID", dbus_machine_id.strpath
        ):
            with mock.patch(
                "uaclient.system.ETC_MACHINE_ID", etc_machine_id.strpath
            ):
                value = system.get_machine_id(cfg)
        assert "dbus-machine-id" == value

    @mock.patch("uaclient.files.state_files.machine_id_file.read")
    def test_get_machine_id_uses_machine_id_from_data_dir(
        self, m_machine_id_file_read, _m_machine_id_file_write, FakeConfig
    ):
        cfg = FakeConfig()
        m_machine_id_file_read.return_value = "data-machine-id"

        with mock.patch("uaclient.util.os.path.exists") as m_exists:
            m_exists.return_value = False
            value = system.get_machine_id(cfg)
        assert "data-machine-id" == value

    @mock.patch("uaclient.files.state_files.machine_id_file.read")
    def test_get_machine_id_create_machine_id_in_data_dir(
        self, m_machine_id_file_read, m_machine_id_file_write, FakeConfig
    ):
        """When no machine-id is found, create one in data_dir using uuid4."""

        cfg = FakeConfig()
        m_machine_id_file_read.return_value = None
        with mock.patch("uaclient.util.os.path.exists") as m_exists:
            with mock.patch("uaclient.system.uuid.uuid4") as m_uuid4:
                m_exists.return_value = False
                m_uuid4.return_value = uuid.UUID(
                    "0123456789abcdef0123456789abcdef"
                )
                value = system.get_machine_id(cfg)

        assert "01234567-89ab-cdef-0123-456789abcdef" == value
        assert [
            mock.call("01234567-89ab-cdef-0123-456789abcdef")
        ] == m_machine_id_file_write.call_args_list

    @pytest.mark.parametrize("empty_value", ["", "\n"])
    @mock.patch("uaclient.files.state_files.machine_id_file.read")
    def test_fallback_used_if_all_other_files_are_empty(
        self,
        m_machine_id_file_read,
        m_machine_id_file_write,
        empty_value,
        fake_machine_token_file,
    ):
        m_machine_id_file_read.return_value = None
        fake_machine_token_file.attached = True
        fake_machine_token_file.token = {"some": "thing"}
        with mock.patch("uaclient.util.os.path.exists") as m_exists:
            m_exists.return_value = True
            with mock.patch(
                "uaclient.system.load_file", return_value=empty_value
            ):
                with mock.patch("uaclient.system.uuid.uuid4") as m_uuid4:
                    m_uuid4.return_value = uuid.UUID(
                        "0123456789abcdef0123456789abcdef"
                    )
                    value = system.get_machine_id.__wrapped__(None)
        assert "01234567-89ab-cdef-0123-456789abcdef" == value
        assert [
            mock.call("01234567-89ab-cdef-0123-456789abcdef")
        ] == m_machine_id_file_write.call_args_list


class TestShouldReboot:
    @pytest.mark.parametrize("path_exists", ((True), (False)))
    @mock.patch("os.path.exists")
    def test_should_reboot_when_no_installed_pkgs_provided(
        self, m_path, path_exists
    ):
        m_path.return_value = path_exists
        assert path_exists == system.should_reboot()
        assert 1 == m_path.call_count

    @mock.patch("os.path.exists")
    @mock.patch("uaclient.system.load_file")
    def test_should_reboot_when_no_reboot_required_pkgs_exist(
        self, m_load_file, m_path
    ):
        installed_pkgs = set(["test"])
        m_path.return_value = True
        m_load_file.side_effect = FileNotFoundError()

        assert system.should_reboot(installed_pkgs)
        assert 1 == m_path.call_count
        assert 1 == m_load_file.call_count

    @pytest.mark.parametrize(
        "installed_pkgs,installed_pkgs_regex,reboot_required_pkgs,"
        "expected_ret",
        (
            (set(["a", "b", "c"]), None, "", False),
            (set(["a", "b", "c"]), None, "a", True),
            (set(["a", "b", "c"]), None, "a\ne", True),
            (set(["a", "b", "c"]), None, "d\ne", False),
            (set(["a", "b", "c"]), set(["t.."]), "a\ne", True),
            (set(["a", "b", "c"]), set(["t.."]), "one\ntwo", True),
            (None, set(["^t..$"]), "one\ntwo", True),
            (None, set(["^t..$"]), "one\nthree", False),
            (None, None, "a\ne", True),
        ),
    )
    @mock.patch("os.path.exists")
    @mock.patch("uaclient.system.load_file")
    def test_should_reboot_when_reboot_required_pkgs_exist(
        self,
        m_load_file,
        m_path,
        installed_pkgs,
        installed_pkgs_regex,
        reboot_required_pkgs,
        expected_ret,
    ):
        m_path.return_value = True
        m_load_file.return_value = reboot_required_pkgs
        assert expected_ret == system.should_reboot(
            installed_pkgs=installed_pkgs,
            installed_pkgs_regex=installed_pkgs_regex,
        )


class TestWriteFile:
    @mock.patch("os.unlink")
    @mock.patch("os.rename")
    @mock.patch("os.makedirs")
    @mock.patch("os.chmod")
    @mock.patch("tempfile.NamedTemporaryFile")
    def test_delete_tempfile_on_error(
        self,
        m_NamedTemporaryFile,
        m_chmod,
        m_makedirs,
        m_rename,
        m_unlink,
    ):
        test_tmpfile = mock.MagicMock()
        test_tmpfile.name = "test_tmpfile"
        m_NamedTemporaryFile.return_value = test_tmpfile

        m_rename.side_effect = Exception()

        with pytest.raises(Exception):
            system.write_file("test", "test")

        assert [mock.call("test_tmpfile")] == m_unlink.call_args_list


class TestSubp:
    def test_raise_error_on_timeout(self, _subp):
        """When cmd exceeds the timeout raises a TimeoutExpired error."""
        with mock.patch("uaclient.system._subp", side_effect=_subp):
            with pytest.raises(subprocess.TimeoutExpired) as excinfo:
                system.subp(["sleep", "2"], timeout=0)
        msg = "Command '[b'sleep', b'2']' timed out after 0 seconds"
        assert msg == str(excinfo.value)

    @mock.patch("uaclient.util.time.sleep")
    def test_default_do_not_retry_on_failure_return_code(self, m_sleep, _subp):
        """When no retry_sleeps are specified, do not retry failures."""
        with mock.patch("uaclient.system._subp", side_effect=_subp):
            with pytest.raises(exceptions.ProcessExecutionError) as excinfo:
                system.subp(["ls", "--bogus"])

        assert 2 == excinfo.value.exit_code
        assert "" == excinfo.value.stdout
        assert 0 == m_sleep.call_count  # no retries

    @mock.patch("uaclient.util.time.sleep")
    def test_no_error_on_accepted_return_codes(self, m_sleep, _subp):
        """When rcs list includes the exit code, do not raise an error."""
        with mock.patch("uaclient.system._subp", side_effect=_subp):
            out, _ = system.subp(["ls", "--bogus"], rcs=[2])

        assert "" == out
        assert 0 == m_sleep.call_count  # no retries

    @mock.patch("uaclient.util.time.sleep")
    def test_retry_with_specified_sleeps_on_error(self, m_sleep, _subp):
        """When retry_sleeps given, use defined sleeps between each retry."""
        with mock.patch("uaclient.system._subp", side_effect=_subp):
            with pytest.raises(exceptions.ProcessExecutionError) as excinfo:
                system.subp(["ls", "--bogus"], retry_sleeps=[1, 3, 0.4])

        expected_error = "Failed running command 'ls --bogus' [exit(2)]"
        assert expected_error in str(excinfo.value)
        expected_sleeps = [mock.call(1), mock.call(3), mock.call(0.4)]
        assert expected_sleeps == m_sleep.call_args_list

    @mock.patch("uaclient.util.time.sleep")
    def test_retry_doesnt_consume_retry_sleeps(self, m_sleep, _subp):
        """When retry_sleeps given, use defined sleeps between each retry."""
        sleeps = [1, 3, 0.4]
        expected_sleeps = sleeps.copy()
        with mock.patch("uaclient.system._subp", side_effect=_subp):
            with pytest.raises(exceptions.ProcessExecutionError):
                system.subp(["ls", "--bogus"], retry_sleeps=sleeps)

        assert expected_sleeps == sleeps

    @pytest.mark.parametrize("caplog_text", [logging.DEBUG], indirect=True)
    @mock.patch("uaclient.system._subp")
    @mock.patch("uaclient.util.time.sleep")
    def test_retry_logs_remaining_retries(self, m_sleep, m_subp, caplog_text):
        """When retry_sleeps given, use defined sleeps between each retry."""
        sleeps = [1, 3, 0.4]
        m_subp.side_effect = exceptions.ProcessExecutionError(
            "Funky apt %d error"
        )
        with pytest.raises(exceptions.ProcessExecutionError):
            system.subp(["apt", "dostuff"], retry_sleeps=sleeps)

        logs = caplog_text()
        expected_logs = [
            "Invalid command specified 'Funky apt %d error'.",
            "Retrying 3 more times.",
            "Invalid command specified 'Funky apt %d error'.",
            "Retrying 2 more times.",
            "Invalid command specified 'Funky apt %d error'.",
            "Retrying 1 more times.",
        ]
        for log in expected_logs:
            assert log in logs

    @pytest.mark.parametrize("caplog_text", [logging.WARNING], indirect=True)
    @pytest.mark.parametrize("capture", [True, False])
    @mock.patch("uaclient.system._subp")
    def test_log_subp_fails_stdout_stderr_capture_toggle(
        self, m_subp, capture, caplog_text
    ):
        """When subp fails, capture the logs in stdout/stderr"""
        out = "Tried downloading file"
        err = "Network error"
        m_subp.side_effect = exceptions.ProcessExecutionError(
            "Serious apt error", stdout=out, stderr=err
        )
        with pytest.raises(exceptions.ProcessExecutionError):
            system.subp(["apt", "nothing"], capture=capture)

        logs = caplog_text()
        expected_logs = ["Stderr: {}".format(err), "Stdout: {}".format(out)]
        for log in expected_logs:
            if capture:
                assert log in logs
            else:
                assert log not in logs

    @pytest.mark.parametrize(
        [
            "override_env_vars",
            "os_environ",
            "expected_env_arg",
        ],
        (
            (None, {}, {"LANG": "C.UTF-8", "LC_ALL": "C.UTF-8"}),
            (
                None,
                {"test": "val"},
                {"test": "val", "LANG": "C.UTF-8", "LC_ALL": "C.UTF-8"},
            ),
            (
                {},
                {"test": "val"},
                {"test": "val", "LANG": "C.UTF-8", "LC_ALL": "C.UTF-8"},
            ),
            (
                {"set": "new"},
                {"test": "val"},
                {
                    "test": "val",
                    "LANG": "C.UTF-8",
                    "LC_ALL": "C.UTF-8",
                    "set": "new",
                },
            ),
            (
                {"set": "new", "test": "newval"},
                {"test": "val"},
                {
                    "LANG": "C.UTF-8",
                    "LC_ALL": "C.UTF-8",
                    "test": "newval",
                    "set": "new",
                },
            ),
        ),
    )
    @mock.patch("subprocess.Popen")
    def test_subp_uses_environment_variables(
        self,
        m_popen,
        override_env_vars,
        os_environ,
        expected_env_arg,
        _subp,
    ):
        mock_process = mock.MagicMock(returncode=0)
        mock_process.communicate.return_value = (b"", b"")
        m_popen.return_value = mock_process

        with mock.patch("os.environ", os_environ):
            _subp(["apt", "nothing"], override_env_vars=override_env_vars)

        assert [
            mock.call(
                [b"apt", b"nothing"],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                env=expected_env_arg,
            )
        ] == m_popen.call_args_list

    @mock.patch("subprocess.Popen")
    def test_subp_no_pipe_stdouterr(
        self,
        m_popen,
        _subp,
    ):
        mock_process = mock.MagicMock(returncode=0)
        mock_process.communicate.return_value = (b"", b"")
        m_popen.return_value = mock_process

        with mock.patch("os.environ", {}):
            _subp(["fake"], pipe_stdouterr=False)

        assert [
            mock.call(
                [b"fake"],
                stdout=None,
                stderr=None,
                env={},
            )
        ] == m_popen.call_args_list

    @mock.patch("subprocess.Popen")
    def test_subp_ignores_non_utf8(
        self,
        m_popen,
        _subp,
    ):
        mock_process = mock.MagicMock(returncode=0)
        mock_process.communicate.return_value = (b"T\x8fE\xacS\x93T\x87!", b"")
        m_popen.return_value = mock_process

        out, _ = _subp(["fake"], pipe_stdouterr=False)

        assert "TEST!" == out


class TestIsSystemdServiceActive:
    @pytest.mark.parametrize(
        [
            "subp_side_effect",
            "expected_return",
        ],
        (
            (None, True),
            (exceptions.ProcessExecutionError("test"), False),
        ),
    )
    @mock.patch("uaclient.system.subp")
    def test_is_systemd_unit_active(
        self,
        m_subp,
        subp_side_effect,
        expected_return,
    ):
        m_subp.side_effect = subp_side_effect
        assert expected_return == system.is_systemd_unit_active("test")


class TestGetSystemdUnitActiveState:
    @pytest.mark.parametrize(
        [
            "subp_side_effect",
            "expected",
        ],
        (
            ([("ActiveState=active\n", "")], "active"),
            ([("ActiveState=activating\n", "")], "activating"),
            ([("", "")], None),
            ([("anything=anything", "")], None),
            (exceptions.ProcessExecutionError("test"), None),
        ),
    )
    @mock.patch("uaclient.system.subp")
    def test_get_systemd_unit_active_state(
        self,
        m_subp,
        subp_side_effect,
        expected,
    ):
        m_subp.side_effect = subp_side_effect
        assert expected == system.get_systemd_unit_active_state("test.service")


class TestGetCpuInfo:
    @pytest.mark.parametrize(
        [
            "cpuinfo_load_side_effect",
            "sysinfo_load_side_effect",
            "devicetree_base_model_load_side_effect",
            "expected_cpuinfo",
        ],
        [
            (
                # x86_64
                (
                    "bogomips        : 4199.88\n"
                    "clflush size    : 64\n"
                    "cache_alignment : 64\n"
                    "address sizes   : 39 bits physical, 48 bits virtual\n"
                    "\n"
                    "processor          : 0\n"
                    "vendor_id          : GenuineIntel\n"
                    "cpu family         : 6\n"
                    "model              : 142\n"
                    "model name         : Test Intel 1\n"
                    "stepping           : 10\n"
                    "microcode          : 0xf4\n"
                    "cpu MHz            : 1600.020\n"
                    "cache size         : 8192 KB\n"
                    "physical id        : 0\n"
                    "siblings           : 8\n"
                    "core id            : 0\n"
                    "cpu cores          : 4\n"
                    "apicid             : 0\n"
                    "initial apicid     : 0\n"
                    "fpu                : yes\n"
                    "fpu_exception      : yes\n"
                    "cpuid level        : 22\n"
                    "wp                 : yes\n"
                    "flags              : fpu vme de pse tsc msr pae mce cx8\n"
                    "\n"
                    "processor          : 1\n"
                    "vendor_id          : GenuineIntel\n"
                    "cpu family         : 6\n"
                    "model              : 142\n"
                    "model name         : Test Intel 2\n"
                ),
                FileNotFoundError(),
                FileNotFoundError(),
                system.CpuInfo(
                    cpuinfo_cpu=None,
                    cpuinfo_cpu_architecture=None,
                    cpuinfo_cpu_family="6",
                    cpuinfo_cpu_implementer=None,
                    cpuinfo_cpu_part=None,
                    cpuinfo_cpu_revision=None,
                    cpuinfo_cpu_variant=None,
                    cpuinfo_model="142",
                    cpuinfo_model_name="Test Intel 1",
                    cpuinfo_stepping="10",
                    cpuinfo_vendor_id="GenuineIntel",
                    sys_firmware_devicetree_base_model=None,
                    sysinfo_model=None,
                    sysinfo_type=None,
                ),
            ),
            (
                # PPC
                (
                    "processor   : 0\n"
                    "cpu         : POWER10\n"
                    "clock       : 2800.000000MHz\n"
                    "revision    : 2.0 (pvr 0080 0200)\n"
                    "\n"
                    "processor   : 1\n"
                    "cpu         : TEST\n"
                    "clock       : 2800.000000MHz\n"
                    "revision    : 2.0 (pvr 0080 0200)\n"
                ),
                FileNotFoundError(),
                FileNotFoundError(),
                system.CpuInfo(
                    cpuinfo_cpu="POWER10",
                    cpuinfo_cpu_architecture=None,
                    cpuinfo_cpu_family=None,
                    cpuinfo_cpu_implementer=None,
                    cpuinfo_cpu_part=None,
                    cpuinfo_cpu_revision=None,
                    cpuinfo_cpu_variant=None,
                    cpuinfo_model=None,
                    cpuinfo_model_name=None,
                    cpuinfo_stepping=None,
                    cpuinfo_vendor_id=None,
                    sys_firmware_devicetree_base_model=None,
                    sysinfo_model=None,
                    sysinfo_type=None,
                ),
            ),
            (
                # S390X
                FileNotFoundError(),
                (
                    "Manufacturer:         IBM\n"
                    "Type:                 2964\n"
                    "Model:                400           N63\n"
                    "Sequence Code:        00000000000B8F67\n"
                    "Plant:                02\n"
                    "Model Capacity:       400           00000000\n"
                    "Capacity Adj. Ind.:   100\n"
                    "Capacity Ch. Reason:  0\n"
                    "Capacity Transient:   0\n"
                    "Type 1 Percentage:    0\n"
                    "Type 2 Percentage:    0\n"
                    "Type 3 Percentage:    0\n"
                    "Type 4 Percentage:    0\n"
                    "Type 5 Percentage:    0\n"
                ),
                FileNotFoundError(),
                system.CpuInfo(
                    cpuinfo_cpu=None,
                    cpuinfo_cpu_architecture=None,
                    cpuinfo_cpu_family=None,
                    cpuinfo_cpu_implementer=None,
                    cpuinfo_cpu_part=None,
                    cpuinfo_cpu_revision=None,
                    cpuinfo_cpu_variant=None,
                    cpuinfo_model=None,
                    cpuinfo_model_name=None,
                    cpuinfo_stepping=None,
                    cpuinfo_vendor_id=None,
                    sys_firmware_devicetree_base_model=None,
                    sysinfo_model="400           N63",
                    sysinfo_type="2964",
                ),
            ),
            (
                # ARM
                (
                    "processor        : 0\n"
                    "model name       : ARMv7 Processor rev 3 (v7l)\n"
                    "BogoMIPS         : 144.00\n"
                    "Features         : half thumb fastmult vfp edsp neon \n"
                    "CPU implementer  : 0x41\n"
                    "CPU architecture : 7\n"
                    "CPU variant      : 0x0\n"
                    "CPU part         : 0xd08\n"
                    "CPU revision     : 3\n"
                ),
                FileNotFoundError(),
                "ARM CPU Type",
                system.CpuInfo(
                    cpuinfo_cpu=None,
                    cpuinfo_cpu_architecture="7",
                    cpuinfo_cpu_family=None,
                    cpuinfo_cpu_implementer="0x41",
                    cpuinfo_cpu_part="0xd08",
                    cpuinfo_cpu_revision="3",
                    cpuinfo_cpu_variant="0x0",
                    cpuinfo_model=None,
                    cpuinfo_model_name="ARMv7 Processor rev 3 (v7l)",
                    cpuinfo_stepping=None,
                    cpuinfo_vendor_id=None,
                    sys_firmware_devicetree_base_model="ARM CPU Type",
                    sysinfo_model=None,
                    sysinfo_type=None,
                ),
            ),
        ],
    )
    @mock.patch("uaclient.system.load_file")
    def test_get_cpu_info(
        self,
        m_load_file,
        cpuinfo_load_side_effect,
        sysinfo_load_side_effect,
        devicetree_base_model_load_side_effect,
        expected_cpuinfo,
    ):
        m_load_file.side_effect = [
            cpuinfo_load_side_effect,
            sysinfo_load_side_effect,
            devicetree_base_model_load_side_effect,
        ]
        assert expected_cpuinfo == system.get_cpu_info.__wrapped__()


class TestGetUserCacheDir:
    @pytest.mark.parametrize(
        [
            "is_root",
            "xdg_cache_home",
            "expanduser_result",
            "expected",
        ],
        (
            (True, None, None, "/run/ubuntu-advantage"),
            (False, None, "/home/user", "/home/user/.cache/ubuntu-pro"),
            (False, "/something", "/home/user", "/something/ubuntu-pro"),
        ),
    )
    @mock.patch("os.path.expanduser")
    @mock.patch("os.environ.get")
    @mock.patch("uaclient.util.we_are_currently_root")
    def test_get_user_cache_dir(
        self,
        m_we_are_currently_root,
        m_environ_get,
        m_expanduser,
        is_root,
        xdg_cache_home,
        expanduser_result,
        expected,
    ):
        m_we_are_currently_root.return_value = is_root
        m_environ_get.return_value = xdg_cache_home
        m_expanduser.return_value = expanduser_result
        assert expected == system.get_user_cache_dir()


class TestGetRebootRequiredPkgs:
    @mock.patch("uaclient.system.load_file")
    def test_when_no_reboot_required_file_is_found(self, m_load_file):
        m_load_file.side_effect = FileNotFoundError()
        assert system.get_reboot_required_pkgs() is None

    @pytest.mark.parametrize(
        "reboot_required_pkgs,expected_standard_pkgs,expected_kernel_pkgs",
        (
            ("", [], []),
            ("pkg1\npkg2", ["pkg1", "pkg2"], []),
            (
                "pkg1\nlinux-image-pkg\npkg2\nlinux-base-pkg",
                ["pkg1", "pkg2"],
                ["linux-base-pkg", "linux-image-pkg"],
            ),
            (
                "linux-image-pkg\nlinux-base-pkg",
                [],
                ["linux-base-pkg", "linux-image-pkg"],
            ),
        ),
    )
    @mock.patch("uaclient.system.load_file")
    def test_reboot_required_pkgs(
        self,
        m_load_file,
        reboot_required_pkgs,
        expected_standard_pkgs,
        expected_kernel_pkgs,
    ):
        m_load_file.return_value = reboot_required_pkgs
        reboot_required_pkgs = system.get_reboot_required_pkgs()

        assert reboot_required_pkgs is not None
        assert expected_standard_pkgs == reboot_required_pkgs.standard_packages
        assert expected_kernel_pkgs == reboot_required_pkgs.kernel_packages


@mock.patch("uaclient.system.util.we_are_currently_root")
class TestGetInstalledUbuntuKernels:
    def test_non_root_user_raises_error(self, m_is_root):
        m_is_root.return_value = False

        with pytest.raises(RuntimeError) as e:
            system.get_installed_ubuntu_kernels()

        assert "needs to be executed as root" in e.value.args[0]

    @pytest.mark.parametrize(
        "installed,files_in_boot,valid_file,expected_return",
        (
            (
                ["linux-image-4.4.0-1020-kvm", "other_package"],
                ["/boot/vmlinuz-4.4.0-1020-kvm"],
                True,
                ["4.4.0-1020-kvm"],
            ),
            (
                ["linux-image-4.4.0-1020-kvm", "other_package"],
                ["/boot/vmlinux-4.4.0-1020-kvm"],
                True,
                ["4.4.0-1020-kvm"],
            ),
            (
                [
                    "linux-image-5.4.0-39-generic",
                    "linux-image-4.15.0-118-generic",
                    "other package",
                ],
                [
                    "/boot/vmlinuz-5.4.0-39-generic",
                    "/boot/vmlinuz-4.15.0-118-generic",
                ],
                True,
                ["5.4.0-39-generic", "4.15.0-118-generic"],
            ),
            (
                ["some package", "other_package"],
                ["/boot/vmlinuz-4.4.0-1020-kvm"],
                True,
                [],
            ),
            (
                ["linux-image-4.4.0-1020-kvm", "other_package"],
                [],
                True,
                [],
            ),
            (
                ["linux-image-4.4.0-1020-kvm", "other_package"],
                ["/boot/vmlinuz-4.4.0-1020-kvm"],
                False,
                [],
            ),
            (
                ["linux-image-5.8.0-63-generic", "linux-image-5.6.0-1056-oem"],
                [
                    "/boot/vmlinuz-5.6.0-1056-oem",
                    "/boot/vmlinuz-5.4.0-77-generic",
                ],
                True,
                ["5.6.0-1056-oem"],
            ),
        ),
    )
    @mock.patch("uaclient.system.subp")
    @mock.patch("uaclient.system.glob.glob")
    @mock.patch("uaclient.apt.get_installed_packages_names")
    def test_only_valid_packages_returned(
        self,
        m_package_names,
        m_glob,
        m_subp,
        m_is_root,
        installed,
        files_in_boot,
        valid_file,
        expected_return,
    ):
        m_is_root.return_value = True
        m_package_names.return_value = installed
        m_glob.return_value = files_in_boot
        m_subp.return_value = (
            ("Some Linux kernel here", False)
            if valid_file
            else ("no good", False)
        )

        assert expected_return == system.get_installed_ubuntu_kernels()



import mock
import pytest

from uaclient.api.u.pro.security.cves._common.v1 import (
    VulnerabilitiesAlreadyFixed,
    VulnerabilityParser,
    VulnerabilityStatus,
    _get_vulnerability_fix_status,
)

M_PATH = "uaclient.api.u.pro.security.cves._common.v1."


VULNERABILITIES_DATA = {
    "packages": {
        "test1": {
            "source_versions": {
                "1.1.2": {
                    "binary_packages": {
                        "test1-bin": "1.1.5",
                        "test1-bin1": "1.1.2",
                    },
                    "pocket": "esm-infra",
                },
                "1.1.3": {
                    "binary_packages": {
                        "test1-bin": "1.1.5",
                        "test1-bin1": "1.1.2",
                        "test1-bin2": "1.1.3",
                    },
                    "pocket": "esm-infra",
                },
                "1.1.4": {
                    "binary_packages": {
                        "test1-bin1": "1.1.5",
                    },
                    "pocket": "esm-infra",
                },
            },
            "cves": {
                "CVE-2022-12345": {
                    "source_fixed_version": None,
                    "status": "vulnerable",
                },
                "CVE-2022-56789": {
                    "source_fixed_version": "1.1.2",
                    "status": "fixed",
                },
                "CVE-2022-18976": {
                    "source_fixed_version": "1.0.9",
                    "status": "vulnerable",
                },
                "CVE-2022-95632": {
                    "source_fixed_version": None,
                    "status": "not-vulnerable",
                },
                "CVE-2022-86782": {
                    "source_fixed_version": "1.1.4",
                    "status": "fixed",
                },
            },
        },
        "test2": {
            "source_versions": {
                "1.1.2": {
                    "binary_packages": {
                        "test2-bin2-1": "1.1.2",
                    },
                    "pocket": "esm-infra",
                }
            },
            "cves": {
                "CVE-2022-12345": {
                    "source_fixed_version": None,
                    "status": "vulnerable",
                },
            },
        },
    },
    "security_issues": {
        "cves": {
            "CVE-2022-12345": {
                "description": "description",
                "published_at": "date",
                "notes": ["hint"],
                "mitigation": "hint",
                "cvss_severity": "low",
                "cvss_score": 4.2,
            },
            "CVE-2022-56789": {
                "description": "description",
                "published_at": "date",
                "notes": ["hint"],
                "mitigation": "hint",
                "cvss_severity": "low",
                "cvss_score": 4.2,
            },
            "CVE-2022-86782": {
                "description": "description",
                "published_at": "date",
                "notes": ["hint"],
                "mitigation": "hint",
                "cvss_severity": "low",
                "cvss_score": 4.2,
            },
        }
    },
}


class ConcreteVulnerabilityParser(VulnerabilityParser):
    vulnerability_type = "cves"

    def get_package_vulnerabilities(self, affected_pkg):
        return affected_pkg.get(self.vulnerability_type, {})

    def _post_process_vulnerability_info(
        self,
        vulnerability_info,
        vulnerabilities_data,
    ):
        return vulnerability_info


class TestVulnerabilityParser:

    @pytest.mark.parametrize(
        "vulnerabilities_data,installed_pkgs_by_source,expected_result",
        (
            (
                VULNERABILITIES_DATA,
                {
                    "test1": {
                        "test1-bin1": "1.1.1",
                        "test1-bin": "1.1.4",
                    }
                },
                {
                    "packages": {
                        "test1-bin1": {
                            "current_version": "1.1.1",
                            "cves": [
                                {
                                    "name": "CVE-2022-12345",
                                    "fix_version": None,
                                    "fix_status": "vulnerable",
                                    "fix_origin": None,
                                },
                                {
                                    "name": "CVE-2022-56789",
                                    "fix_version": "1.1.2",
                                    "fix_status": "fixed",
                                    "fix_origin": "esm-infra",
                                },
                                {
                                    "name": "CVE-2022-86782",
                                    "fix_version": "1.1.5",
                                    "fix_status": "fixed",
                                    "fix_origin": "esm-infra",
                                },
                            ],
                        },
                        "test1-bin": {
                            "current_version": "1.1.4",
                            "cves": [
                                {
                                    "name": "CVE-2022-12345",
                                    "fix_version": None,
                                    "fix_status": "vulnerable",
                                    "fix_origin": None,
                                },
                                {
                                    "name": "CVE-2022-56789",
                                    "fix_version": "1.1.5",
                                    "fix_status": "fixed",
                                    "fix_origin": "esm-infra",
                                },
                                {
                                    "name": "CVE-2022-86782",
                                    "fix_version": None,
                                    "fix_status": "unknown",
                                    "fix_origin": None,
                                },
                            ],
                        },
                    },
                    "vulnerabilities": {
                        "CVE-2022-12345": {
                            "description": "description",
                            "published_at": "date",
                            "notes": ["hint"],
                            "mitigation": "hint",
                            "cvss_severity": "low",
                            "cvss_score": 4.2,
                        },
                        "CVE-2022-56789": {
                            "description": "description",
                            "published_at": "date",
                            "notes": ["hint"],
                            "mitigation": "hint",
                            "cvss_severity": "low",
                            "cvss_score": 4.2,
                        },
                        "CVE-2022-86782": {
                            "description": "description",
                            "published_at": "date",
                            "notes": ["hint"],
                            "mitigation": "hint",
                            "cvss_severity": "low",
                            "cvss_score": 4.2,
                        },
                    },
                },
            ),
        ),
    )
    @mock.patch(
        M_PATH + "VulnerabilityParser._get_installed_source_pkg_version"
    )
    def test_get_vulnerabilities_for_installed_pkgs(
        self,
        m_get_installed_source_pkg_version,
        vulnerabilities_data,
        installed_pkgs_by_source,
        expected_result,
    ):
        m_get_installed_source_pkg_version.return_value = "1.1.3"
        parser = ConcreteVulnerabilityParser()
        assert (
            parser.get_vulnerabilities_for_installed_pkgs(
                vulnerabilities_data, installed_pkgs_by_source
            ).vulnerabilities_info
            == expected_result
        )


class TestGetVulnerabilityFixStatus:
    @pytest.mark.parametrize(
        "affected_pkgs,expected_state",
        (
            (
                [
                    {
                        "fix_version": "1.0",
                        "name": "pkg1",
                    },
                    {
                        "fix_version": "1.0",
                        "name": "pkg2",
                    },
                ],
                VulnerabilityStatus.FULL_FIX_AVAILABLE,
            ),
            (
                [
                    {
                        "fix_version": None,
                        "name": "pkg1",
                    },
                    {
                        "fix_version": "1.0",
                        "name": "pkg2",
                    },
                ],
                VulnerabilityStatus.PARTIAL_FIX_AVAILABLE,
            ),
            (
                [
                    {
                        "fix_version": None,
                        "name": "pkg1",
                    },
                    {
                        "fix_version": None,
                        "name": "pkg2",
                    },
                ],
                VulnerabilityStatus.NO_FIX_AVAILABLE,
            ),
        ),
    )
    def test_get_vulnerability_fix_status(
        self,
        affected_pkgs,
        expected_state,
    ):
        assert _get_vulnerability_fix_status(affected_pkgs) == expected_state


class TestVulnerabilitiesAlreadyFixed:

    @pytest.mark.parametrize(
        "vulnerabilities,expected_result",
        (
            (
                [
                    ("CVE-1234-5", "ubuntu_pro", "high"),
                    ("CVE-1234-5", "ubuntu_pro", "high"),
                    ("CVE-1234-5", "ubuntu_security", "high"),
                    ("CVE-1234-5", "ubuntu_security", "high"),
                    ("CVE-1552-5", "ubuntu_pro", "medium"),
                    ("CVE-1382-5", "ubuntu_security", "low"),
                ],
                {
                    "count": {
                        "ubuntu_pro": 2,
                        "ubuntu_security": 2,
                    },
                    "info": {
                        "ubuntu_pro": {"high": 1, "medium": 1},
                        "ubuntu_security": {"high": 1, "low": 1},
                    },
                },
            ),
            (
                [
                    ("USN-1234-5", "ubuntu_pro"),
                    ("USN-1234-5", "ubuntu_security"),
                    ("USN-1552-5", "ubuntu_pro"),
                    ("USN-1382-5", "ubuntu_security"),
                ],
                {
                    "count": {
                        "ubuntu_pro": 2,
                        "ubuntu_security": 2,
                    },
                    "info": {
                        "ubuntu_pro": {},
                        "ubuntu_security": {},
                    },
                },
            ),
        ),
    )
    def test_vulnerabilities_already_fixed_to_dict(
        self,
        vulnerabilities,
        expected_result,
    ):
        vulnerabilities_already_fixed = VulnerabilitiesAlreadyFixed()
        for vulnerability in vulnerabilities:
            vulnerabilities_already_fixed.add_vulnerability(*vulnerability)

        assert expected_result == vulnerabilities_already_fixed.to_dict()



#!/usr/bin/env python3

"""
This script is called after running do-release-upgrade in a machine.
This converts list files to deb822 files when upgrading to Noble.
"""

import logging
import os
import sys

from aptsources.sourceslist import SourceEntry  # type: ignore

from uaclient import defaults, entitlements
from uaclient.apt import _get_sources_file_content
from uaclient.config import UAConfig
from uaclient.log import setup_cli_logging
from uaclient.system import (
    ensure_file_absent,
    get_release_info,
    load_file,
    write_file,
)
from uaclient.util import set_filename_extension

if __name__ == "__main__":
    series = get_release_info().series
    if series != "noble":
        sys.exit(0)

    setup_cli_logging(logging.DEBUG, defaults.CONFIG_DEFAULTS["log_file"])
    cfg = UAConfig()

    for entitlement_class in entitlements.ENTITLEMENT_CLASSES:
        if not issubclass(
            entitlement_class, entitlements.repo.RepoEntitlement
        ):
            continue

        entitlement = entitlement_class(cfg)

        filename = set_filename_extension(entitlement.repo_file, "list")
        if os.path.exists(filename):
            # If do-release-upgrade commented out the file, whether the
            # repository is not reachable or is considered a third party, then
            # it will be handled in upgrade_lts_contract. This script only
            # changes services which are enabled, active and reachable.
            valid_sources = [
                SourceEntry(line)
                for line in load_file(filename).strip().split("\n")
                if line.strip().startswith("deb")
            ]
            if valid_sources:
                # get this information from the file, to avoid interacting with
                # the entitlement_config
                suites = list(set(source.dist for source in valid_sources))
                repo_url = valid_sources[0].uri
                include_deb_src = any(
                    source.type == "deb-src" for source in valid_sources
                )
                content = _get_sources_file_content(
                    suites,
                    series,
                    True,
                    repo_url,
                    entitlement.repo_key_file,
                    include_deb_src,
                )
                write_file(entitlement.repo_file, content)

            ensure_file_absent(filename)



import mock

try:  # Drop try-except after xenial EOL
    from contextlib import AbstractContextManager
except ImportError:
    AbstractContextManager = object


class does_not_raise(AbstractContextManager):
    """Reentrant noop context manager.
    Useful to parametrize tests raising and not raising exceptions.

    Note: In python-3.7+, this can be substituted by contextlib.nullcontext
    More info:
    https://docs.pytest.org/en/6.2.x/example/parametrize.html?highlight=does_not_raise#parametrizing-conditional-raising

    Example:
    --------
    >>> @pytest.mark.parametrize(
    >>>     "example_input,expectation",
    >>>     [
    >>>         (1, does_not_raise()),
    >>>         (0, pytest.raises(ZeroDivisionError)),
    >>>     ],
    >>> )
    >>> def test_division(example_input, expectation):
    >>>     with expectation:
    >>>         assert (0 / example_input) is not None
    """

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass


def mock_with_name_attr(*args, name, mock_name=None, **kwargs):
    m = mock.MagicMock(*args, name=mock_name, **kwargs)
    m.name = name
    return m



from typing import Tuple, Type

from uaclient import messages
from uaclient.entitlements import repo
from uaclient.entitlements.base import EntitlementWithMessage, UAEntitlement


class ROSCommonEntitlement(repo.RepoEntitlement):
    help_doc_url = messages.urls.ROS_HOME_PAGE
    repo_key_file = "ubuntu-pro-ros.gpg"


class ROSEntitlement(ROSCommonEntitlement):
    name = "ros"
    title = messages.ROS_TITLE
    description = messages.ROS_DESCRIPTION
    help_text = messages.ROS_HELP_TEXT
    origin = "UbuntuROS"

    @property
    def required_services(self) -> Tuple[EntitlementWithMessage, ...]:
        from uaclient.entitlements.esm import (
            ESMAppsEntitlement,
            ESMInfraEntitlement,
        )

        return (
            EntitlementWithMessage(
                ESMInfraEntitlement,
                messages.ROS_REQUIRES_ESM,
            ),
            EntitlementWithMessage(
                ESMAppsEntitlement,
                messages.ROS_REQUIRES_ESM,
            ),
        )

    @property
    def dependent_services(self) -> Tuple[Type[UAEntitlement], ...]:
        return (ROSUpdatesEntitlement,)


class ROSUpdatesEntitlement(ROSCommonEntitlement):
    name = "ros-updates"
    title = messages.ROS_UPDATES_TITLE
    description = messages.ROS_UPDATES_DESCRIPTION
    help_text = messages.ROS_UPDATES_HELP_TEXT
    origin = "UbuntuROSUpdates"

    @property
    def required_services(self) -> Tuple[EntitlementWithMessage, ...]:
        from uaclient.entitlements.esm import (
            ESMAppsEntitlement,
            ESMInfraEntitlement,
        )

        return (
            EntitlementWithMessage(
                ESMInfraEntitlement,
                messages.ROS_REQUIRES_ESM,
            ),
            EntitlementWithMessage(
                ESMAppsEntitlement,
                messages.ROS_REQUIRES_ESM,
            ),
            EntitlementWithMessage(
                ROSEntitlement,
                messages.ROS_UPDATES_REQUIRES_ROS,
            ),
        )



import datetime

import pytest

from uaclient.api.u.pro.status.is_attached.v1 import (
    ContractExpiryStatus,
    _is_attached,
)


class TestGetContractExpiryStatus:
    @pytest.mark.parametrize(
        "contract_remaining_days,expected_status",
        (
            (21, ContractExpiryStatus.ACTIVE),
            (20, ContractExpiryStatus.ACTIVE_EXPIRED_SOON),
            (-1, ContractExpiryStatus.EXPIRED_GRACE_PERIOD),
            (-20, ContractExpiryStatus.EXPIRED),
        ),
    )
    def test_contract_expiry_status_based_on_remaining_days(
        self, contract_remaining_days, expected_status, fake_machine_token_file
    ):
        """Return a tuple of ContractExpiryStatus and remaining_days"""
        now = datetime.datetime.utcnow()
        expire_date = now + datetime.timedelta(days=contract_remaining_days)
        fake_machine_token_file.token = {
            "machineTokenInfo": {"contractInfo": {"effectiveTo": expire_date}}
        }
        is_attached_info = _is_attached(None)
        assert is_attached_info.is_attached
        assert expected_status.value == is_attached_info.contract_status
        assert (
            contract_remaining_days == is_attached_info.contract_remaining_days
        )



import mock
import pytest

from uaclient import exceptions, http
from uaclient.clouds.lxd import LXDAutoAttachInstance
from uaclient.testing import helpers

M_PATH = "uaclient.clouds.lxd."


class TestLXDAutoAttachInstance:
    def test_is_viable(self):
        instance = LXDAutoAttachInstance()
        assert instance.is_viable

    def should_poll_for_pro_license(self):
        instance = LXDAutoAttachInstance()
        assert instance.should_poll_for_pro_license()

    @pytest.mark.parametrize(
        [
            "wait_for_change",
            "response",
            "expected_raises",
            "expected_result",
        ],
        [
            (
                True,
                http.HTTPResponse(
                    code=200, headers={}, body="", json_dict={}, json_list=[]
                ),
                pytest.raises(exceptions.CancelProLicensePolling),
                None,
            ),
            (
                False,
                http.HTTPResponse(
                    code=400, headers={}, body="", json_dict={}, json_list=[]
                ),
                helpers.does_not_raise(),
                False,
            ),
            (
                False,
                http.HTTPResponse(
                    code=200, headers={}, body="", json_dict={}, json_list=[]
                ),
                helpers.does_not_raise(),
                False,
            ),
            (
                False,
                http.HTTPResponse(
                    code=200,
                    headers={},
                    body="",
                    json_dict={"guest_attach": "off"},
                    json_list=[],
                ),
                helpers.does_not_raise(),
                False,
            ),
            (
                False,
                http.HTTPResponse(
                    code=200,
                    headers={},
                    body="",
                    json_dict={"guest_attach": "on"},
                    json_list=[],
                ),
                helpers.does_not_raise(),
                True,
            ),
            (
                False,
                http.HTTPResponse(
                    code=200,
                    headers={},
                    body="",
                    json_dict={"guest_attach": "available"},
                    json_list=[],
                ),
                helpers.does_not_raise(),
                False,
            ),
        ],
    )
    @mock.patch(M_PATH + "http.unix_socket_request")
    def test_is_pro_license_present(
        self,
        m_unix_socket_request,
        wait_for_change,
        response,
        expected_raises,
        expected_result,
    ):
        m_unix_socket_request.return_value = response
        instance = LXDAutoAttachInstance()
        with expected_raises:
            assert expected_result == instance.is_pro_license_present(
                wait_for_change=wait_for_change
            )

    @pytest.mark.parametrize(
        [
            "response",
            "expected_raises",
            "expected_result",
        ],
        [
            (
                http.HTTPResponse(
                    code=404, headers={}, body="", json_dict={}, json_list=[]
                ),
                pytest.raises(exceptions.LXDAutoAttachNotAvailable),
                None,
            ),
            (
                http.HTTPResponse(
                    code=403, headers={}, body="", json_dict={}, json_list=[]
                ),
                pytest.raises(exceptions.LXDAutoAttachNotAllowed),
                None,
            ),
            (
                http.HTTPResponse(
                    code=200,
                    headers={},
                    body="",
                    json_dict={"guest_token": "token"},
                    json_list=[],
                ),
                helpers.does_not_raise(),
                "token",
            ),
        ],
    )
    @mock.patch(M_PATH + "http.unix_socket_request")
    def test_acquire_pro_token(
        self,
        m_unix_socket_request,
        response,
        expected_raises,
        expected_result,
        FakeConfig,
    ):
        m_unix_socket_request.return_value = response
        instance = LXDAutoAttachInstance()
        with expected_raises:
            assert expected_result == instance.acquire_pro_token(FakeConfig())



#!/usr/bin/env python3

import logging

from uaclient import log
from uaclient.apt import update_esm_caches
from uaclient.config import UAConfig

LOG = logging.getLogger("ubuntupro.lib.esm_cache")


def main(cfg: UAConfig) -> None:
    try:
        update_esm_caches(cfg)
    except Exception as e:
        msg = getattr(e, "msg", str(e))
        LOG.error("Error updating the cache: %s", msg)


if __name__ == "__main__":
    log.setup_journald_logging()
    cfg = UAConfig()
    main(cfg)



from uaclient.files.data_types import DataObjectFile, DataObjectFileFormat
from uaclient.files.files import UAFile
from uaclient.files.machine_token import MachineTokenFile
from uaclient.files.user_config_file import UserConfigFileObject

__all__ = [
    "DataObjectFile",
    "DataObjectFileFormat",
    "MachineTokenFile",
    "UAFile",
    "UserConfigFileObject",
]



import abc
import json
import os
import posixpath
from typing import Any, Dict, Optional  # noqa: F401
from urllib.parse import urlencode

from uaclient import config, http, system, util, version


class UAServiceClient(metaclass=abc.ABCMeta):

    url_timeout = 30  # type: Optional[int]
    # Cached serviceclient_url_responses if provided in uaclient.conf
    # via features: {serviceclient_url_responses: /some/file.json}
    _response_overlay = None  # type: Dict[str, Any]

    @property
    @abc.abstractmethod
    def cfg_url_base_attr(self) -> str:
        """String in subclasses, the UAConfig attribute containing base url"""
        pass

    def __init__(self, cfg: Optional[config.UAConfig] = None) -> None:
        if not cfg:
            self.cfg = config.UAConfig()
        else:
            self.cfg = cfg

    def headers(self):
        return {
            "user-agent": "UA-Client/{}".format(version.get_version()),
            "accept": "application/json",
            "content-type": "application/json",
        }

    def request_url(
        self,
        path,
        data=None,
        headers=None,
        method=None,
        query_params=None,
        log_response_body: bool = True,
        timeout: Optional[int] = None,
    ) -> http.HTTPResponse:
        path = path.lstrip("/")
        if not headers:
            headers = self.headers()
        if headers.get("content-type") == "application/json" and data:
            data = json.dumps(data, cls=util.DatetimeAwareJSONEncoder).encode(
                "utf-8"
            )
        url = posixpath.join(getattr(self.cfg, self.cfg_url_base_attr), path)

        fake_response = self._get_fake_responses(url)
        if fake_response:
            return fake_response  # URL faked by uaclient.conf

        if query_params:
            # filter out None values
            filtered_params = {
                k: v for k, v in sorted(query_params.items()) if v is not None
            }
            url += "?" + urlencode(filtered_params)
        timeout_to_use = timeout if timeout is not None else self.url_timeout

        return http.readurl(
            url=url,
            data=data,
            headers=headers,
            method=method,
            timeout=timeout_to_use,
            log_response_body=log_response_body,
        )

    def _get_response_overlay(self, url: str):
        """Return a list of fake response dicts for a given URL.

        serviceclient_url_responses in uaclient.conf should be a path
        to a json file which contains a dictionary keyed by full URL path.
        Each value will be a list of dicts representing each faked response
        for the given URL.

            The response dict item will have a code: <HTTP_STATUS_CODE> and
               response: "some string of content".
            The JSON string below fakes the available_resources URL on the
            contract server:
            '{"https://contracts.canonical.com/v1/resources": \
               [{"code": 200, "response": {"key": "val1", "key2": "val2"}}]}'

        :return: List of dicts for each faked response matching the url, or
           and empty list when no matching url found.
        """
        if self._response_overlay is not None:
            # Cache it so we don't re-read config every readurl call
            return self._response_overlay.get(url, [])
        response_overlay_path = self.cfg.features.get(
            "serviceclient_url_responses"
        )
        if not response_overlay_path:
            self._response_overlay = {}
        elif not os.path.exists(response_overlay_path):
            self._response_overlay = {}
        else:
            self._response_overlay = json.loads(
                system.load_file(response_overlay_path)
            )
        return self._response_overlay.get(url, [])

    def _get_fake_responses(self, url: str) -> Optional[http.HTTPResponse]:
        """Return response if faked for this URL in uaclient.conf."""
        responses = self._get_response_overlay(url)
        if not responses:
            return None

        if len(responses) == 1:
            # When only one response is defined, repeat it for all calls
            response = responses[0]
        else:
            # When multiple responses defined pop the first one off the list.
            response = responses.pop(0)

        json_dict = {}
        json_list = []
        resp = response["response"]
        if isinstance(resp, dict):
            json_dict = resp
        elif isinstance(resp, list):
            json_list = resp

        return http.HTTPResponse(
            code=response["code"],
            headers=response.get("headers", {}),
            body=json.dumps(response["response"], sort_keys=True),
            json_dict=json_dict,
            json_list=json_list,
        )



import contextlib
import io
import json
import textwrap

import mock
import pytest

from uaclient import entitlements, event_logger, exceptions, lock, messages
from uaclient.api.u.pro.services.dependencies.v1 import (
    ServiceWithDependencies,
    ServiceWithReason,
)
from uaclient.cli import main_error_handler
from uaclient.cli.disable import (
    disable_command,
    prompt_for_dependency_handling,
)
from uaclient.entitlements.entitlement_status import (
    CanDisableFailure,
    CanDisableFailureReason,
)
from uaclient.testing.helpers import does_not_raise


@pytest.fixture
def all_service_msg(FakeConfig):
    ALL_SERVICE_MSG = "\n".join(
        textwrap.wrap(
            "Try "
            + ", ".join(entitlements.valid_services(cfg=FakeConfig()))
            + ".",
            width=80,
            break_long_words=False,
            break_on_hyphens=False,
        )
    )
    return ALL_SERVICE_MSG


class TestDisable:
    @pytest.mark.parametrize("service", [["testitlement"], ["ent1", "ent2"]])
    @pytest.mark.parametrize("assume_yes", (True, False))
    @pytest.mark.parametrize(
        "disable_return,return_code", ((True, 0), (False, 1))
    )
    @mock.patch("uaclient.cli.disable._enabled_services")
    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    @mock.patch(
        "uaclient.cli.disable.contract.UAContractClient.update_activity_token",
    )
    @mock.patch("uaclient.entitlements.entitlement_factory")
    @mock.patch("uaclient.entitlements.valid_services")
    @mock.patch("uaclient.status.status")
    def test_entitlement_instantiated_and_disabled(
        self,
        m_status,
        m_valid_services,
        m_entitlement_factory,
        m_update_activity_token,
        _m_check_lock_info,
        m_enabled_services,
        disable_return,
        return_code,
        assume_yes,
        service,
        tmpdir,
        capsys,
        event,
        FakeConfig,
        fake_machine_token_file,
    ):
        entitlements_obj = []
        ent_dict = {}
        m_valid_services.return_value = []

        if not disable_return:
            fail = CanDisableFailure(
                CanDisableFailureReason.ALREADY_DISABLED,
                message=messages.NamedMessage("test-code", "test"),
            )
        else:
            fail = None

        for entitlement_name in service:
            m_entitlement = mock.MagicMock()
            m_entitlement._check_for_reboot.return_value = False
            m_entitlement.enabled_variant = None
            m_entitlement.disable.return_value = (disable_return, fail)

            entitlements_obj.append(m_entitlement)
            m_valid_services.return_value.append(entitlement_name)
            type(m_entitlement).name = mock.PropertyMock(
                return_value=entitlement_name
            )
            ent_dict[entitlement_name] = m_entitlement

        m_enabled_services.return_value = mock.MagicMock(
            enabled_services=ent_dict.values()
        )

        def factory_side_effect(cfg, name, **kwargs):
            return ent_dict.get(name, mock.MagicMock())

        m_entitlement_factory.side_effect = factory_side_effect

        cfg = FakeConfig()
        fake_machine_token_file.attached = True
        args_mock = mock.Mock()
        args_mock.service = service
        args_mock.assume_yes = assume_yes
        args_mock.purge = False

        with mock.patch.object(lock, "lock_data_file"):
            ret = disable_command.action(args_mock, cfg=cfg)

        expected_disable_call = mock.call(mock.ANY)
        for m_entitlement in entitlements_obj:
            assert [
                expected_disable_call
            ] == m_entitlement.disable.call_args_list

        assert return_code == ret
        assert len(entitlements_obj) == m_status.call_count
        assert 1 == m_update_activity_token.call_count

        cfg = FakeConfig()
        args_mock.assume_yes = True
        args_mock.format = "json"
        with mock.patch.object(
            event, "_event_logger_mode", event_logger.EventLoggerMode.JSON
        ):
            with mock.patch.object(event, "set_event_mode"):
                with mock.patch.object(lock, "lock_data_file"):
                    fake_stdout = io.StringIO()
                    with contextlib.redirect_stdout(fake_stdout):
                        ret = disable_command.action(args_mock, cfg=cfg)

        expected = {
            "_schema_version": event_logger.JSON_SCHEMA_VERSION,
            "result": "success" if disable_return else "failure",
            "errors": [],
            "failed_services": [] if disable_return else service,
            "needs_reboot": False,
            "processed_services": service if disable_return else [],
            "warnings": [],
        }

        if not disable_return:
            expected["errors"] = [
                {
                    "message": "test",
                    "message_code": "test-code",
                    "service": ent_name,
                    "type": "service",
                }
                for ent_name in service
            ]

        assert return_code == ret
        assert expected == json.loads(fake_stdout.getvalue())

    @pytest.mark.parametrize("assume_yes", (True, False))
    @mock.patch("uaclient.cli.disable._enabled_services")
    @mock.patch("uaclient.contract.UAContractClient.update_activity_token")
    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    @mock.patch("uaclient.entitlements.entitlement_factory")
    @mock.patch("uaclient.entitlements.valid_services")
    @mock.patch("uaclient.status.status")
    def test_entitlements_not_found_disabled_and_enabled(
        self,
        m_status,
        m_valid_services,
        m_entitlement_factory,
        _m_check_lock_info,
        _m_update_activity_token,
        m_enabled_services,
        assume_yes,
        tmpdir,
        event,
        FakeConfig,
        fake_machine_token_file,
    ):
        expected_error_tmpl = messages.E_INVALID_SERVICE_OP_FAILURE
        num_calls = 2

        m_ent1_obj = mock.MagicMock()
        m_ent1_obj.enabled_variant = None
        m_ent1_obj.disable.return_value = (
            False,
            CanDisableFailure(
                CanDisableFailureReason.ALREADY_DISABLED,
                message=messages.NamedMessage("test-code", "test"),
            ),
        )
        type(m_ent1_obj).name = mock.PropertyMock(return_value="ent1")
        m_ent1_obj._check_for_reboot.return_value = False

        m_ent2_obj = mock.MagicMock()
        m_ent2_obj.enabled_variant = None
        m_ent2_obj.disable.return_value = (
            False,
            CanDisableFailure(
                CanDisableFailureReason.ALREADY_DISABLED,
                message=messages.NamedMessage("test-code2", "test2"),
            ),
        )
        type(m_ent2_obj).name = mock.PropertyMock(return_value="ent2")
        m_ent2_obj._check_for_reboot.return_value = False

        m_ent3_cls = mock.MagicMock()
        m_ent3_obj = m_ent3_cls.return_value
        m_ent3_obj.enabled_variant = None
        m_ent3_obj.disable.return_value = (True, None)
        type(m_ent3_obj).name = mock.PropertyMock(return_value="ent3")
        m_ent3_obj._check_for_reboot.return_value = False

        def factory_side_effect(cfg, name, **kwargs):
            if name == "ent2":
                return m_ent2_obj
            if name == "ent3":
                return m_ent3_obj
            return mock.MagicMock()

        m_entitlement_factory.side_effect = factory_side_effect
        m_valid_services.return_value = ["ent2", "ent3"]
        m_enabled_services.return_value = mock.MagicMock(
            enabled_service=[m_ent1_obj, m_ent2_obj, m_ent3_obj]
        )

        cfg = FakeConfig()
        fake_machine_token_file.attached = True
        args_mock = mock.Mock()
        args_mock.service = ["ent1", "ent2", "ent3"]
        args_mock.assume_yes = assume_yes
        args_mock.purge = False

        first_fake_stdout = io.StringIO()
        with contextlib.redirect_stdout(first_fake_stdout):
            with mock.patch.object(lock, "lock_data_file"):
                disable_command.action(args_mock, cfg=cfg)

        assert (
            expected_error_tmpl.format(
                operation="disable",
                invalid_service="ent1",
                service_msg="Try ent2, ent3.",
            ).msg
            in first_fake_stdout.getvalue()
        )

        expected_disable_call = mock.call(mock.ANY)
        for m_ent in [m_ent2_obj, m_ent3_obj]:
            assert [expected_disable_call] == m_ent.disable.call_args_list

        assert 0 == m_ent1_obj.call_count
        assert num_calls == m_status.call_count

        args_mock.assume_yes = True
        args_mock.format = "json"
        with mock.patch.object(lock, "lock_data_file"):
            fake_stdout = io.StringIO()
            with contextlib.redirect_stdout(fake_stdout):
                main_error_handler(disable_command.action)(args_mock, cfg=cfg)

        expected = {
            "_schema_version": event_logger.JSON_SCHEMA_VERSION,
            "result": "failure",
            "errors": [
                {
                    "message": "test2",
                    "message_code": "test-code2",
                    "service": "ent2",
                    "type": "service",
                },
                {
                    "additional_info": {
                        "invalid_service": "ent1",
                        "operation": "disable",
                        "service_msg": "Try ent2, ent3.",
                    },
                    "message": (
                        "Cannot disable unknown service 'ent1'.\n"
                        "Try ent2, ent3."
                    ),
                    "message_code": "invalid-service-or-failure",
                    "service": None,
                    "type": "system",
                },
            ],
            "failed_services": ["ent2"],
            "needs_reboot": False,
            "processed_services": ["ent3"],
            "warnings": [],
        }

        assert expected == json.loads(fake_stdout.getvalue())

    @pytest.mark.parametrize(
        "root,expected_error_template",
        [
            (False, messages.E_NONROOT_USER),
        ],
    )
    @mock.patch("uaclient.contract.UAContractClient.update_activity_token")
    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    @mock.patch("uaclient.util.we_are_currently_root")
    def test_invalid_service_error_message(
        self,
        m_we_are_currently_root,
        _m_check_lock_info,
        _m_update_activity_token,
        root,
        expected_error_template,
        FakeConfig,
        fake_machine_token_file,
        event,
        all_service_msg,
    ):
        """Check invalid service name results in custom error message."""
        m_we_are_currently_root.return_value = root
        fake_machine_token_file.attached = True

        cfg = FakeConfig()
        args = mock.MagicMock()
        args.purge = False
        args.service = ["esm-infra"]

        expected_error = expected_error_template
        expected_info = None

        with pytest.raises(exceptions.UbuntuProError) as err:
            with mock.patch.object(lock, "lock_data_file"):
                disable_command.action(args, cfg)

        assert expected_error.msg == err.value.msg

        args.assume_yes = True
        args.format = "json"
        with pytest.raises(SystemExit):
            with mock.patch.object(
                event, "_event_logger_mode", event_logger.EventLoggerMode.JSON
            ):
                with mock.patch.object(event, "set_event_mode"):
                    fake_stdout = io.StringIO()
                    with contextlib.redirect_stdout(fake_stdout):
                        with mock.patch.object(lock, "lock_data_file"):
                            main_error_handler(disable_command.action)(
                                args, cfg
                            )

        expected = {
            "_schema_version": event_logger.JSON_SCHEMA_VERSION,
            "result": "failure",
            "errors": [
                {
                    "message": expected_error.msg,
                    "message_code": expected_error.name,
                    "service": None,
                    "type": "system",
                }
            ],
            "failed_services": [],
            "needs_reboot": False,
            "processed_services": [],
            "warnings": [],
        }
        if expected_info is not None:
            expected["errors"][0]["additional_info"] = expected_info
        assert expected == json.loads(fake_stdout.getvalue())

    @pytest.mark.parametrize("service", [["bogus"], ["bogus1", "bogus2"]])
    @mock.patch("uaclient.contract.UAContractClient.update_activity_token")
    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    def test_invalid_service_names(
        self,
        _m_check_lock_info,
        _m_update_activity_token,
        service,
        FakeConfig,
        fake_machine_token_file,
        event,
        all_service_msg,
    ):
        expected_error_tmpl = messages.E_INVALID_SERVICE_OP_FAILURE

        cfg = FakeConfig()
        fake_machine_token_file.attached = True
        args = mock.MagicMock()
        args.purge = False
        expected_error = expected_error_tmpl.format(
            operation="disable",
            invalid_service=", ".join(sorted(service)),
            service_msg=all_service_msg,
        )
        first_fake_stdout = io.StringIO()
        with contextlib.redirect_stdout(first_fake_stdout):
            with mock.patch.object(lock, "lock_data_file"):
                args.service = service
                disable_command.action(args, cfg)

        assert expected_error.msg in first_fake_stdout.getvalue()

        args.assume_yes = True
        args.format = "json"
        with mock.patch.object(lock, "lock_data_file"):
            fake_stdout = io.StringIO()
            with contextlib.redirect_stdout(fake_stdout):
                main_error_handler(disable_command.action)(args, cfg)

        expected = {
            "_schema_version": event_logger.JSON_SCHEMA_VERSION,
            "result": "failure",
            "errors": [
                {
                    "additional_info": {
                        "invalid_service": ", ".join(sorted(service)),
                        "operation": "disable",
                        "service_msg": all_service_msg,
                    },
                    "message": expected_error.msg,
                    "message_code": expected_error.name,
                    "service": None,
                    "type": "system",
                }
            ],
            "failed_services": [],
            "needs_reboot": False,
            "processed_services": [],
            "warnings": [],
        }
        assert expected == json.loads(fake_stdout.getvalue())

    @pytest.mark.parametrize(
        "root,expected_error_template",
        [
            (True, messages.E_VALID_SERVICE_FAILURE_UNATTACHED),
            (False, messages.E_NONROOT_USER),
        ],
    )
    @mock.patch("uaclient.util.we_are_currently_root")
    def test_unattached_error_message(
        self,
        m_we_are_currently_root,
        root,
        expected_error_template,
        FakeConfig,
        event,
    ):
        """Check that root user gets unattached message."""
        m_we_are_currently_root.return_value = root

        cfg = FakeConfig()
        args = mock.MagicMock()
        args.command = "disable"
        if root:
            expected_error = expected_error_template.format(
                valid_service="esm-infra", operation="disable"
            )
            expected_info = {
                "valid_service": "esm-infra",
                "operation": "disable",
            }
        else:
            expected_error = expected_error_template
            expected_info = None

        with pytest.raises(exceptions.UbuntuProError) as err:
            args.service = ["esm-infra"]
            disable_command.action(args, cfg)

        assert expected_error.msg == err.value.msg

        args.assume_yes = True
        args.format = "json"
        with pytest.raises(SystemExit):
            with mock.patch.object(
                event, "_event_logger_mode", event_logger.EventLoggerMode.JSON
            ):
                with mock.patch.object(event, "set_event_mode"):
                    fake_stdout = io.StringIO()
                    with contextlib.redirect_stdout(fake_stdout):
                        main_error_handler(disable_command.action)(args, cfg)

        expected = {
            "_schema_version": event_logger.JSON_SCHEMA_VERSION,
            "result": "failure",
            "errors": [
                {
                    "message": expected_error.msg,
                    "message_code": expected_error.name,
                    "service": None,
                    "type": "system",
                }
            ],
            "failed_services": [],
            "needs_reboot": False,
            "processed_services": [],
            "warnings": [],
        }
        if expected_info is not None:
            expected["errors"][0]["additional_info"] = expected_info
        assert expected == json.loads(fake_stdout.getvalue())

    @mock.patch("uaclient.lock.check_lock_info")
    @mock.patch("time.sleep")
    @mock.patch("uaclient.system.subp")
    def test_lock_file_exists(
        self,
        m_subp,
        m_sleep,
        m_check_lock_info,
        FakeConfig,
        fake_machine_token_file,
        event,
    ):
        """Check inability to disable if operation in progress holds lock."""
        cfg = FakeConfig()
        fake_machine_token_file.attached = True
        args = mock.MagicMock()
        expected_error = messages.E_LOCK_HELD_ERROR.format(
            lock_request="pro disable", lock_holder="pro enable", pid="123"
        )
        m_check_lock_info.return_value = (123, "pro enable")
        with pytest.raises(exceptions.LockHeldError) as err:
            args.service = ["esm-infra"]
            disable_command.action(args, cfg)
        assert 12 == m_check_lock_info.call_count
        assert expected_error.msg == err.value.msg

        args.assume_yes = True
        args.format = "json"
        with pytest.raises(SystemExit):
            with mock.patch.object(
                event, "_event_logger_mode", event_logger.EventLoggerMode.JSON
            ):
                with mock.patch.object(event, "set_event_mode"):
                    fake_stdout = io.StringIO()
                    with contextlib.redirect_stdout(fake_stdout):
                        main_error_handler(disable_command.action)(args, cfg)

        expected = {
            "_schema_version": event_logger.JSON_SCHEMA_VERSION,
            "result": "failure",
            "errors": [
                {
                    "additional_info": {
                        "lock_holder": "pro enable",
                        "lock_request": "pro disable",
                        "pid": 123,
                    },
                    "message": expected_error.msg,
                    "message_code": expected_error.name,
                    "service": None,
                    "type": "system",
                }
            ],
            "failed_services": [],
            "needs_reboot": False,
            "processed_services": [],
            "warnings": [],
        }
        assert expected == json.loads(fake_stdout.getvalue())

    def test_format_json_fails_when_assume_yes_flag_not_used(self, event):
        cfg = mock.MagicMock()
        args_mock = mock.MagicMock()
        args_mock.format = "json"
        args_mock.assume_yes = False

        with pytest.raises(SystemExit):
            with mock.patch.object(
                event, "_event_logger_mode", event_logger.EventLoggerMode.JSON
            ):
                fake_stdout = io.StringIO()
                with contextlib.redirect_stdout(fake_stdout):
                    main_error_handler(disable_command.action)(args_mock, cfg)

        expected_message = messages.E_JSON_FORMAT_REQUIRE_ASSUME_YES
        expected = {
            "_schema_version": event_logger.JSON_SCHEMA_VERSION,
            "result": "failure",
            "errors": [
                {
                    "message": expected_message.msg,
                    "message_code": expected_message.name,
                    "service": None,
                    "type": "system",
                }
            ],
            "failed_services": [],
            "needs_reboot": False,
            "processed_services": [],
            "warnings": [],
        }
        assert expected == json.loads(fake_stdout.getvalue())

    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    @mock.patch("uaclient.cli.cli_util._is_attached")
    def test_purge_assume_yes_incompatible(
        self, m_is_attached, _m_check_lock_info, capsys
    ):
        cfg = mock.MagicMock()
        args_mock = mock.MagicMock()
        args_mock.service = "test"
        args_mock.assume_yes = True
        args_mock.purge = True

        m_is_attached.return_value = mock.MagicMock(
            is_attached=True,
            contract_status="active",
            contract_remaining_days=100,
        )

        with pytest.raises(SystemExit):
            with mock.patch.object(lock, "lock_data_file"):
                main_error_handler(disable_command.action)(args_mock, cfg)

        _out, err = capsys.readouterr()

        assert (
            messages.E_INVALID_OPTION_COMBINATION.format(
                option1="--purge", option2="--assume-yes"
            ).msg
            in err.strip()
        )


class TestPromptForDependencyHandling:
    @pytest.mark.parametrize(
        [
            "service",
            "all_dependencies",
            "enabled_service_names",
            "called_name",
            "service_title",
            "prompt_side_effects",
            "expected_prompts",
            "expected_raise",
        ],
        [
            # no dependencies
            (
                "one",
                [
                    ServiceWithDependencies(
                        name="two", incompatible_with=[], depends_on=[]
                    )
                ],
                [],
                "one",
                "One",
                [],
                [],
                does_not_raise(),
            ),
            # required by "two", but two not enabled
            (
                "one",
                [
                    ServiceWithDependencies(
                        name="two",
                        incompatible_with=[],
                        depends_on=[
                            ServiceWithReason(
                                name="one", reason=mock.MagicMock()
                            )
                        ],
                    )
                ],
                [],
                "one",
                "One",
                [],
                [],
                does_not_raise(),
            ),
            # required by "two", two enabled, successful prompt
            (
                "one",
                [
                    ServiceWithDependencies(
                        name="two",
                        incompatible_with=[],
                        depends_on=[
                            ServiceWithReason(
                                name="one", reason=mock.MagicMock()
                            )
                        ],
                    )
                ],
                ["two"],
                "one",
                "One",
                [True],
                [mock.call(msg=mock.ANY)],
                does_not_raise(),
            ),
            # required by "two", two enabled, denied prompt
            (
                "one",
                [
                    ServiceWithDependencies(
                        name="two",
                        incompatible_with=[],
                        depends_on=[
                            ServiceWithReason(
                                name="one", reason=mock.MagicMock()
                            )
                        ],
                    )
                ],
                ["two"],
                "one",
                "One",
                [False],
                [mock.call(msg=mock.ANY)],
                pytest.raises(exceptions.DependentServiceStopsDisable),
            ),
            # required by "two" and "three", three enabled, success
            (
                "one",
                [
                    ServiceWithDependencies(
                        name="two",
                        incompatible_with=[],
                        depends_on=[
                            ServiceWithReason(
                                name="one", reason=mock.MagicMock()
                            )
                        ],
                    ),
                    ServiceWithDependencies(
                        name="three",
                        incompatible_with=[],
                        depends_on=[
                            ServiceWithReason(
                                name="one", reason=mock.MagicMock()
                            )
                        ],
                    ),
                ],
                ["three"],
                "one",
                "One",
                [True],
                [mock.call(msg=mock.ANY)],
                does_not_raise(),
            ),
        ],
    )
    @mock.patch("uaclient.entitlements.get_title")
    @mock.patch("uaclient.util.prompt_for_confirmation")
    def test_prompt_for_dependency_handling(
        self,
        m_prompt_for_confirmation,
        m_entitlement_get_title,
        service,
        all_dependencies,
        enabled_service_names,
        called_name,
        service_title,
        prompt_side_effects,
        expected_prompts,
        expected_raise,
        FakeConfig,
    ):
        m_entitlement_get_title.side_effect = (
            lambda cfg, name, variant="": name.title()
        )
        m_prompt_for_confirmation.side_effect = prompt_side_effects

        with expected_raise:
            prompt_for_dependency_handling(
                FakeConfig(),
                service,
                all_dependencies,
                enabled_service_names,
                called_name,
                service_title,
            )

        assert expected_prompts == m_prompt_for_confirmation.call_args_list



import json

from uaclient import messages, security_status, util
from uaclient.cli.commands import (
    ProArgument,
    ProArgumentGroup,
    ProArgumentMutuallyExclusiveGroup,
    ProCommand,
)
from uaclient.cli.parser import HelpCategory
from uaclient.yaml import safe_dump


def action_security_status(args, *, cfg, **kwargs):
    if args.format == "text":
        if args.thirdparty:
            security_status.list_third_party_packages()
        elif args.unavailable:
            security_status.list_unavailable_packages()
        elif args.esm_infra:
            security_status.list_esm_infra_packages(cfg)
        elif args.esm_apps:
            security_status.list_esm_apps_packages(cfg)
        else:
            security_status.security_status(cfg)
    elif args.format == "json":
        print(
            json.dumps(
                security_status.security_status_dict(cfg),
                sort_keys=True,
                cls=util.DatetimeAwareJSONEncoder,
            )
        )
    else:
        print(
            safe_dump(
                security_status.security_status_dict(cfg),
                default_flow_style=False,
            )
        )
    return 0


security_status_command = ProCommand(
    "security-status",
    help=messages.CLI_ROOT_SECURITY_STATUS,
    description=messages.CLI_SS_DESC,
    preserve_description=True,
    action=action_security_status,
    help_category=HelpCategory.QUICKSTART,
    help_position=5,
    argument_groups=[
        ProArgumentGroup(
            arguments=[
                ProArgument(
                    "--format",
                    help=messages.CLI_FORMAT_DESC.format(default="text"),
                    choices=("json", "yaml", "text"),
                    default="text",
                )
            ],
            mutually_exclusive_groups=[
                ProArgumentMutuallyExclusiveGroup(
                    arguments=[
                        ProArgument(
                            "--thirdparty",
                            help=messages.CLI_SS_THIRDPARTY,
                            action="store_true",
                        ),
                        ProArgument(
                            "--unavailable",
                            help=messages.CLI_SS_UNAVAILABLE,
                            action="store_true",
                        ),
                        ProArgument(
                            "--esm-infra",
                            help=messages.CLI_SS_ESM_INFRA,
                            action="store_true",
                        ),
                        ProArgument(
                            "--esm-apps",
                            help=messages.CLI_SS_ESM_APPS,
                            action="store_true",
                        ),
                    ]
                )
            ],
        )
    ],
)



import logging

import mock
import pytest

from uaclient import exceptions, http
from uaclient.clouds.azure import IMDS_BASE_URL, AzureAutoAttachInstance

M_PATH = "uaclient.clouds.azure."


class TestAzureAutoAttachInstance:
    def test_cloud_type(self):
        """cloud_type is returned as azure."""
        instance = AzureAutoAttachInstance()
        assert "azure" == instance.cloud_type

    @mock.patch(M_PATH + "http.readurl")
    def test_identity_doc_from_azure_url_pkcs7(self, readurl):
        """Return attested signature and compute info as Azure identity doc"""

        def fake_readurl(url, headers, timeout):
            if "attested/document" in url:
                return http.HTTPResponse(
                    code=200,
                    headers={"header": "stuff"},
                    body="",
                    json_dict={"signature": "attestedWOOT!==="},
                    json_list=[],
                )
            elif "instance/compute" in url:
                return http.HTTPResponse(
                    code=200,
                    headers={"header": "stuff"},
                    body="",
                    json_dict={"computekey": "computeval"},
                    json_list=[],
                )
            else:
                raise AssertionError("Unexpected URL provided %s" % url)

        readurl.side_effect = fake_readurl
        instance = AzureAutoAttachInstance()
        assert {
            "compute": {"computekey": "computeval"},
            "pkcs7": "attestedWOOT!===",
        } == instance.identity_doc
        url1 = IMDS_BASE_URL + "instance/compute?api-version=2020-09-01"
        url2 = IMDS_BASE_URL + "attested/document?api-version=2020-09-01"
        assert [
            mock.call(url1, headers={"Metadata": "true"}, timeout=1),
            mock.call(url2, headers={"Metadata": "true"}, timeout=1),
        ] == readurl.call_args_list

    @pytest.mark.parametrize("caplog_text", [logging.DEBUG], indirect=True)
    @pytest.mark.parametrize("fail_count,exception", ((3, False), (4, True)))
    @mock.patch(M_PATH + "util.time.sleep")
    @mock.patch(M_PATH + "http.readurl")
    def test_retry_backoff_on_failed_identity_doc(
        self, readurl, sleep, fail_count, exception, caplog_text
    ):
        """Retries backoff before failing to get Azure.identity_doc"""

        def fake_someurlerrors(url, headers, timeout):
            if readurl.call_count <= fail_count:
                return http.HTTPResponse(
                    code=700 + readurl.call_count,
                    headers={},
                    body="funky error msg",
                    json_dict={},
                    json_list=[],
                )
            if "attested" in url:
                return http.HTTPResponse(
                    code=200,
                    headers={"header": "stuff"},
                    body="",
                    json_dict={"signature": "attestedWOOT!==="},
                    json_list=[],
                )
            elif "compute" in url:
                return http.HTTPResponse(
                    code=200,
                    headers={"header": "stuff"},
                    body="",
                    json_dict={"computekey": "computeval"},
                    json_list=[],
                )
            raise AssertionError("Unexpected url requested {}".format(url))

        readurl.side_effect = fake_someurlerrors
        instance = AzureAutoAttachInstance()
        if exception:
            with pytest.raises(exceptions.CloudMetadataError):
                instance.identity_doc
        else:
            assert {
                "pkcs7": "attestedWOOT!===",
                "compute": {"computekey": "computeval"},
            } == instance.identity_doc

        expected_sleep_calls = [mock.call(1), mock.call(1), mock.call(1)]
        assert expected_sleep_calls == sleep.call_args_list
        expected_logs = [
            "An error occurred while talking the the cloud metadata service: 701 - funky error msg: Retrying 3 more times.",  # noqa: E501
            "An error occurred while talking the the cloud metadata service: 702 - funky error msg: Retrying 2 more times.",  # noqa: E501
            "An error occurred while talking the the cloud metadata service: 703 - funky error msg: Retrying 1 more times.",  # noqa: E501
        ]
        logs = caplog_text()
        for log in expected_logs:
            assert log in logs

    @pytest.mark.parametrize(
        "chassis_asset_tag,ovf_env_exists,viable",
        (
            (None, True, True),
            ("7783-7084-3265-9085-8269-3286-77", False, True),
            ("6783-7084-3265-9085-8269-3286-77", True, True),
            (None, False, False),
            ("6783-7084-3265-9085-8269-3286-77", False, False),
        ),
    )
    @mock.patch(M_PATH + "os.path.exists")
    @mock.patch(M_PATH + "system.load_file")
    def test_is_viable_based_on_dmi_chassis_asset_tag_or_ovf_env(
        self, load_file, m_exists, chassis_asset_tag, ovf_env_exists, viable
    ):
        """Platform viable if chassis asset tag matches or ovf.env exists."""

        def fake_exists(f_name):
            if f_name == "/sys/class/dmi/id/chassis_asset_tag":
                return bool(chassis_asset_tag is not None)
            elif f_name == "/var/lib/cloud/seed/azure/ovf-env.xml":
                return ovf_env_exists
            raise AssertionError("Invalid os.path.exist of {}".format(f_name))

        m_exists.side_effect = fake_exists

        def fake_load_file(f_name):
            if f_name == "/sys/class/dmi/id/chassis_asset_tag":
                return chassis_asset_tag
            raise AssertionError("Invalid load_file of {}".format(f_name))

        load_file.side_effect = fake_load_file
        instance = AzureAutoAttachInstance()
        assert viable is instance.is_viable

    def test_should_poll_for_license(self):
        instance = AzureAutoAttachInstance()
        result = instance.should_poll_for_pro_license()
        assert result

    @pytest.mark.parametrize(
        "metadata_response, expected_result",
        (
            ({}, False),
            ({"licenseType": None}, False),
            ({"licenseType": ""}, False),
            ({"licenseType": "RHEL_BYOS"}, False),
            ({"licenseType": "SLES_BYOS"}, False),
            ({"licenseType": "UBUNTU_PRO"}, True),
        ),
    )
    @mock.patch(M_PATH + "http.readurl")
    def test_is_license_present(
        self, m_readurl, metadata_response, expected_result
    ):
        instance = AzureAutoAttachInstance()
        m_readurl.return_value = http.HTTPResponse(
            code=200,
            headers={},
            body="",
            json_dict=metadata_response,
            json_list=[],
        )
        result = instance.is_pro_license_present(wait_for_change=False)
        assert expected_result == result

    def test_is_licence_present_wait_for_change_raises_exception(self):
        instance = AzureAutoAttachInstance()
        with pytest.raises(exceptions.CancelProLicensePolling):
            instance.is_pro_license_present(wait_for_change=True)



import copy
import datetime

import mock
import pytest

from uaclient.api.u.pro.security.cves.v1 import (
    AffectedPackage,
    CVEAffectedPackage,
    CVEInfo,
    CVEsOptions,
    CVEsResult,
    RelatedUSN,
    _cves,
)

M_PATH = "uaclient.api.u.pro.security.cves.v1."
M_VULN_COMMON_PATH = "uaclient.api.u.pro.security.cves._common.v1."

VULNEBILITIES_DATA = {
    "published_at": "2024-06-24T13:19:16",
    "packages": {
        "test1": {
            "source_versions": {
                "1.1.2": {
                    "binary_packages": {
                        "test1-bin": "1.1.2",
                        "test1-bin1": "1.1.2",
                    },
                    "pocket": "esm-infra",
                }
            },
            "cves": {
                "CVE-2022-12345": {
                    "source_fixed_version": None,
                    "status": "vulnerable",
                },
                "CVE-2022-56789": {
                    "source_fixed_version": "1.1.2",
                    "status": "fixed",
                },
                "CVE-2022-18976": {
                    "source_fixed_version": "1.0.9",
                    "status": "vulnerable",
                },
                "CVE-2022-95632": {
                    "source_fixed_version": None,
                    "status": "not-vulnerable",
                },
            },
        }
    },
    "security_issues": {
        "cves": {
            "CVE-2022-12345": {
                "description": "description",
                "published_at": "2024-06-24T13:19:16",
                "ubuntu_priority": "low",
                "notes": ["hint"],
                "mitigation": "hint",
                "cvss_severity": "low",
                "cvss_score": 4.2,
                "related_usns": ["USN-1582-1"],
                "related_packages": ["test1-bin", "test1-bin1"],
            },
            "CVE-2022-56789": {
                "description": "description",
                "published_at": "2024-06-24T13:19:16",
                "ubuntu_priority": "low",
                "notes": ["hint"],
                "mitigation": "hint",
                "cvss_severity": "low",
                "cvss_score": 4.2,
                "related_usns": [],
            },
        },
        "usns": {
            "USN-1582-1": {
                "title": "USN Title",
                "description": "description",
            },
        },
    },
}


INSTALLED_PKGS_BY_SOURCE = {
    "test1": {
        "test1-bin1": "1.1.1",
        "test1-bin": "1.1.1",
    }
}


class TestCVEs:
    @pytest.mark.parametrize(
        "vulnerabilities_data,installed_pkgs_by_source,cve_options,expected_result",  # noqa
        (
            (
                VULNEBILITIES_DATA,
                INSTALLED_PKGS_BY_SOURCE,
                CVEsOptions(),
                CVEsResult(
                    packages={
                        "test1-bin": AffectedPackage(
                            current_version="1.1.1",
                            cves=[
                                CVEAffectedPackage(
                                    name="CVE-2022-12345",
                                    fix_version=None,
                                    fix_status="vulnerable",
                                    fix_origin=None,
                                ),
                                CVEAffectedPackage(
                                    name="CVE-2022-56789",
                                    fix_version="1.1.2",
                                    fix_status="fixed",
                                    fix_origin="esm-infra",
                                ),
                            ],
                        ),
                        "test1-bin1": AffectedPackage(
                            current_version="1.1.1",
                            cves=[
                                CVEAffectedPackage(
                                    name="CVE-2022-12345",
                                    fix_version=None,
                                    fix_status="vulnerable",
                                    fix_origin=None,
                                ),
                                CVEAffectedPackage(
                                    name="CVE-2022-56789",
                                    fix_version="1.1.2",
                                    fix_status="fixed",
                                    fix_origin="esm-infra",
                                ),
                            ],
                        ),
                    },
                    cves={
                        "CVE-2022-12345": CVEInfo(
                            description="description",
                            priority="low",
                            notes=["hint"],
                            published_at=datetime.datetime(
                                2024,
                                6,
                                24,
                                13,
                                19,
                                16,
                                tzinfo=datetime.timezone.utc,
                            ),
                            cvss_severity="low",
                            cvss_score=4.2,
                            related_usns=[
                                RelatedUSN(
                                    name="USN-1582-1",
                                    title="USN Title",
                                )
                            ],
                            related_packages=["test1-bin", "test1-bin1"],
                        ),
                        "CVE-2022-56789": CVEInfo(
                            description="description",
                            priority="low",
                            notes=["hint"],
                            published_at=datetime.datetime(
                                2024,
                                6,
                                24,
                                13,
                                19,
                                16,
                                tzinfo=datetime.timezone.utc,
                            ),
                            cvss_severity="low",
                            cvss_score=4.2,
                            related_usns=[],
                            related_packages=[],
                        ),
                    },
                    vulnerability_data_published_at=datetime.datetime(
                        2024, 6, 24, 13, 19, 16, tzinfo=datetime.timezone.utc
                    ),
                    apt_updated_at=datetime.datetime(2024, 6, 24, 13, 19, 16),
                ),
            ),
            (
                VULNEBILITIES_DATA,
                INSTALLED_PKGS_BY_SOURCE,
                CVEsOptions(unfixable=True),
                CVEsResult(
                    packages={
                        "test1-bin": AffectedPackage(
                            current_version="1.1.1",
                            cves=[
                                CVEAffectedPackage(
                                    name="CVE-2022-12345",
                                    fix_version=None,
                                    fix_status="vulnerable",
                                    fix_origin=None,
                                ),
                            ],
                        ),
                        "test1-bin1": AffectedPackage(
                            current_version="1.1.1",
                            cves=[
                                CVEAffectedPackage(
                                    name="CVE-2022-12345",
                                    fix_version=None,
                                    fix_status="vulnerable",
                                    fix_origin=None,
                                ),
                            ],
                        ),
                    },
                    cves={
                        "CVE-2022-12345": CVEInfo(
                            description="description",
                            priority="low",
                            notes=["hint"],
                            published_at=datetime.datetime(
                                2024,
                                6,
                                24,
                                13,
                                19,
                                16,
                                tzinfo=datetime.timezone.utc,
                            ),
                            cvss_severity="low",
                            cvss_score=4.2,
                            related_usns=[
                                RelatedUSN(
                                    name="USN-1582-1",
                                    title="USN Title",
                                )
                            ],
                            related_packages=["test1-bin", "test1-bin1"],
                        ),
                    },
                    vulnerability_data_published_at=datetime.datetime(
                        2024, 6, 24, 13, 19, 16, tzinfo=datetime.timezone.utc
                    ),
                    apt_updated_at=datetime.datetime(2024, 6, 24, 13, 19, 16),
                ),
            ),
            (
                VULNEBILITIES_DATA,
                INSTALLED_PKGS_BY_SOURCE,
                CVEsOptions(fixable=True),
                CVEsResult(
                    packages={
                        "test1-bin": AffectedPackage(
                            current_version="1.1.1",
                            cves=[
                                CVEAffectedPackage(
                                    name="CVE-2022-56789",
                                    fix_version="1.1.2",
                                    fix_status="fixed",
                                    fix_origin="esm-infra",
                                ),
                            ],
                        ),
                        "test1-bin1": AffectedPackage(
                            current_version="1.1.1",
                            cves=[
                                CVEAffectedPackage(
                                    name="CVE-2022-56789",
                                    fix_version="1.1.2",
                                    fix_status="fixed",
                                    fix_origin="esm-infra",
                                ),
                            ],
                        ),
                    },
                    cves={
                        "CVE-2022-56789": CVEInfo(
                            description="description",
                            priority="low",
                            notes=["hint"],
                            published_at=datetime.datetime(
                                2024,
                                6,
                                24,
                                13,
                                19,
                                16,
                                tzinfo=datetime.timezone.utc,
                            ),
                            cvss_severity="low",
                            cvss_score=4.2,
                            related_usns=[],
                            related_packages=[],
                        ),
                    },
                    vulnerability_data_published_at=datetime.datetime(
                        2024, 6, 24, 13, 19, 16, tzinfo=datetime.timezone.utc
                    ),
                    apt_updated_at=datetime.datetime(2024, 6, 24, 13, 19, 16),
                ),
            ),
        ),
    )
    @mock.patch(
        M_VULN_COMMON_PATH + "VulnerabilityResultCache.save_result_cache"
    )
    @mock.patch(M_PATH + "get_apt_cache_datetime")
    @mock.patch(M_VULN_COMMON_PATH + "VulnerabilityData.get")
    @mock.patch(
        M_VULN_COMMON_PATH + "VulnerabilityData.refreshed",
        new_callable=mock.PropertyMock,
    )
    @mock.patch(M_VULN_COMMON_PATH + "query_installed_source_pkg_versions")
    def test_parse_data(
        self,
        m_get_source_pkgs,
        m_vulnerability_data_refreshed,
        m_vulnerability_data_get,
        m_get_apt_cache_datetime,
        _m_vulnerability_result_save_cache,
        vulnerabilities_data,
        installed_pkgs_by_source,
        cve_options,
        expected_result,
    ):
        m_vulnerability_data_refreshed.return_vale = True
        m_get_source_pkgs.return_value = installed_pkgs_by_source
        m_vulnerability_data_get.return_value = copy.deepcopy(
            vulnerabilities_data
        )
        m_get_apt_cache_datetime.return_value = datetime.datetime(
            2024, 6, 24, 13, 19, 16
        )
        assert _cves(cve_options, None) == expected_result



import datetime
import glob
import hashlib
import hmac
import json
import logging
import os
import re
import shutil
import subprocess
import sys
import tempfile
import time
from base64 import b64encode
from dataclasses import dataclass
from enum import Enum
from typing import Callable, Iterable, List, Optional, Tuple
from urllib.parse import quote
from urllib.request import Request, urlopen

import yaml

from uaclient.system import get_dpkg_arch

SUT = "sut"  # system under test
BUILDER_NAME_PREFIX = "builder-"
LXC_PROPERTY_MAP = {
    "image": {"series": "properties.release", "machine_type": "Type"},
    "container": {"series": "image.release", "machine_type": "image.type"},
}
UA_TMP_DIR = os.path.join(tempfile.gettempdir(), "uaclient-behave")
SOURCE_PR_TGZ = os.path.join(UA_TMP_DIR, "pr_source.tar.gz")
SOURCE_PR_UNTAR_DIR = os.path.join(UA_TMP_DIR, "behave-ua-src")
SBUILD_DIR = os.path.join(UA_TMP_DIR, "sbuild")
UA_DEB_BUILD_CACHE = os.path.join(UA_TMP_DIR, "deb-cache")


ALL_BINARY_PACKAGE_NAMES = [
    "ubuntu-pro-client",
    "ubuntu-pro-client-l10n",
    "ubuntu-pro-auto-attach",
    "ubuntu-advantage-tools",
    "ubuntu-advantage-pro",
]
NORMAL_BINARY_PACKAGE_NAMES = [
    "ubuntu-pro-client",
    "ubuntu-pro-client-l10n",
    "ubuntu-advantage-tools",
]


@dataclass
class ProDebPaths:
    ubuntu_pro_client: str
    ubuntu_pro_image_auto_attach: str
    ubuntu_pro_client_l10n: str
    ubuntu_advantage_tools: str
    ubuntu_advantage_pro: str

    def non_cloud_pro_image_debs(self, series: str) -> List[Tuple[str, str]]:
        if series in ("xenial", "bionic", "focal", "jammy"):
            return [
                ("ubuntu-pro-client", self.ubuntu_pro_client),
                ("ubuntu-advantage-tools", self.ubuntu_advantage_tools),
                ("ubuntu-pro-client-l10n", self.ubuntu_pro_client_l10n),
            ]
        else:
            return [
                ("ubuntu-pro-client", self.ubuntu_pro_client),
                ("ubuntu-pro-client-l10n", self.ubuntu_pro_client_l10n),
            ]

    def cloud_pro_image_debs(self) -> List[Tuple[str, str]]:
        return [
            (
                "ubuntu-pro-auto-attach",
                self.ubuntu_pro_image_auto_attach,
            ),
            ("ubuntu-advantage-pro", self.ubuntu_advantage_pro),
        ]

    def all_debs(self, series: str) -> List[Tuple[str, str]]:
        return (
            self.non_cloud_pro_image_debs(series) + self.cloud_pro_image_debs()
        )


class InstallationSource(Enum):
    ARCHIVE = "archive"
    PREBUILT = "prebuilt"
    LOCAL = "local"
    DAILY = "daily"
    STAGING = "staging"
    STABLE = "stable"
    PROPOSED = "proposed"
    CUSTOM = "custom"


def repo_state_hash(
    exclude_dirs: Iterable[str] = (
        ".github",
        "demo",
        "features",
        "sru",
        "tools",
        "docs",
        "dev-docs",
        "keyrings",
    )
):
    """
    Generate a string that represents the current local state of
    the git repository.

    This takes (1) the latest commit hash, which represents the committed
    state, (2) the `git diff`, which includes all local changes, and (3)
    the contents of new files not yet committed, and concatenates them.
    Then it generates the md5 hash of the concatenated result.

    By doing so, the result is a unique-enough string that shouldn't
    collide with any other local state in our lifetimes. We can then
    include this string in the name of a build and reuse that build if
    the repo state doesn't change.

    :param exclude_dirs: When non-empty, we exclude all folders in the list
        from the git diff and from the new_file_content. With the default
        value, we don't include any changes to integration tests or other
        unrelated scripts and files in the git repo when generating our local
        state hash. This is useful when iterating on an integration test or
        manual test script because you won't have to re-sbuild if you don't
        change the actual source code of the package.

    :return: A unique string representing the current local state of the
        git repository
    """
    git_rev_cmd = ["git", "rev-parse", "HEAD"]
    rev_out = subprocess.check_output(git_rev_cmd)

    git_diff_cmd = ["git", "diff"]
    exclude_set = set(exclude_dirs)
    if len(exclude_set) > 0:
        files_and_folders_set = set(os.listdir("."))
        filtered_files_and_folders = files_and_folders_set - exclude_set
        git_diff_cmd += ["--", *filtered_files_and_folders]
    diff_out = subprocess.check_output(git_diff_cmd)

    git_status_cmd = ["git", "status", "--porcelain"]
    status_out = subprocess.check_output(git_status_cmd).decode("utf-8")
    status_lines = status_out.split("\n")
    new_files = [name[3:] for name in status_lines if name.startswith("?? ")]
    new_file_content = ""
    for fname in new_files:
        exclude = False
        for exclude_dir in exclude_dirs:
            if fname.startswith(exclude_dir):
                exclude = True
                break
        if exclude:
            continue
        if os.path.isdir(fname):
            continue
        with open(fname) as f:
            new_file_content += f.read()

    output_to_hash = rev_out + diff_out + new_file_content.encode("utf-8")
    return hashlib.md5(output_to_hash).hexdigest()


def get_debs_for_series(debs_path: str, series: str) -> ProDebPaths:
    ubuntu_pro_client = ""
    ubuntu_pro_client_l10n = ""
    ubuntu_pro_image_auto_attach = ""
    ubuntu_advantage_tools = ""
    ubuntu_advantage_pro = ""
    for deb_file in os.listdir(debs_path):
        if series in deb_file:
            full_path = os.path.join(debs_path, deb_file)
            if "ubuntu-pro-client-l10n" in deb_file:
                ubuntu_pro_client_l10n = full_path
            elif "ubuntu-pro-client" in deb_file:
                ubuntu_pro_client = full_path
            elif "ubuntu-pro-auto-attach" in deb_file:
                ubuntu_pro_image_auto_attach = full_path
            elif "ubuntu-advantage-tools" in deb_file:
                ubuntu_advantage_tools = full_path
            elif "ubuntu-advantage-pro" in deb_file:
                ubuntu_advantage_pro = full_path
    return ProDebPaths(
        ubuntu_pro_client=ubuntu_pro_client,
        ubuntu_pro_client_l10n=ubuntu_pro_client_l10n,
        ubuntu_pro_image_auto_attach=ubuntu_pro_image_auto_attach,
        ubuntu_advantage_tools=ubuntu_advantage_tools,
        ubuntu_advantage_pro=ubuntu_advantage_pro,
    )


def _create_deb_path(prefix: str, name: str):
    return os.path.join(UA_DEB_BUILD_CACHE, "{}{}.deb".format(prefix, name))


def build_debs(
    series: str,
    architecture: Optional[str] = None,
    chroot: Optional[str] = None,
    sbuild_output_to_terminal: bool = False,
) -> ProDebPaths:
    """
    Build the package through sbuild and store the debs into
    output_deb_dir

    :param series: The target series to build the package for
    :param output_deb_dir: the target directory in which to copy deb artifacts

    :return: A list of file paths to debs created by the build.
    """
    if architecture is None:
        architecture = get_dpkg_arch()

    deb_prefix = "{}-{}-{}-".format(series, architecture, repo_state_hash())
    deb_paths = ProDebPaths(
        ubuntu_pro_client=_create_deb_path(deb_prefix, "ubuntu-pro-client"),
        ubuntu_pro_image_auto_attach=_create_deb_path(
            deb_prefix, "ubuntu-pro-auto-attach"
        ),
        ubuntu_pro_client_l10n=_create_deb_path(
            deb_prefix, "ubuntu-pro-client-l10n"
        ),
        ubuntu_advantage_tools=_create_deb_path(
            deb_prefix, "ubuntu-advantage-tools"
        ),
        ubuntu_advantage_pro=_create_deb_path(
            deb_prefix, "ubuntu-advantage-pro"
        ),
    )

    if not os.path.exists(UA_DEB_BUILD_CACHE):
        os.makedirs(UA_DEB_BUILD_CACHE)

    if os.path.exists(deb_paths.ubuntu_pro_client):
        logging.info("--- Using debs in cache")
        return deb_paths

    logging.info("--- Creating: {}".format(SOURCE_PR_TGZ))

    cwd = os.getcwd()
    os.chdir("..")
    subprocess.run(
        [
            "tar",
            "-zcf",
            SOURCE_PR_TGZ,
            "--exclude-vcs",
            "--exclude-vcs-ignores",
            os.path.basename(cwd),
        ],
        check=True,
    )
    os.chdir(cwd)

    logging.info("--- Creating: {}".format(SOURCE_PR_UNTAR_DIR))

    # Delete cached folder for pro code
    if os.path.exists(SOURCE_PR_UNTAR_DIR):
        shutil.rmtree(SOURCE_PR_UNTAR_DIR)

    os.makedirs(SOURCE_PR_UNTAR_DIR)
    subprocess.run(
        [
            "tar",
            "-xvf",
            SOURCE_PR_TGZ,
            "-C",
            SOURCE_PR_UNTAR_DIR,
            "--strip-components=1",
        ],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        check=True,
    )

    if os.path.exists(SBUILD_DIR):
        shutil.rmtree(SBUILD_DIR)
    os.makedirs(SBUILD_DIR)

    sbuild_cmd = [
        "sbuild",
        "--chroot-mode",
        "schroot",
        "--no-run-lintian",
        "--resolve-alternatives",
        "--no-clean-source",
        "--build-dir",
        SBUILD_DIR,
        "--arch",
        architecture,
        "-d",
        series,
        SOURCE_PR_UNTAR_DIR,
    ]
    if chroot is not None:
        sbuild_cmd += ["--chroot", chroot]
    else:
        # use ua-series-arch chroot if present
        ua_chroot = "ua-{}-{}".format(series, architecture)
        proc = subprocess.run(
            ["schroot", "--info", "--chroot", ua_chroot],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        if proc.returncode == 0:
            sbuild_cmd += ["--chroot", ua_chroot]

    # disable unit-test during sbuild
    env = os.environ.copy()
    env["DEB_BUILD_OPTIONS"] = env.get("DEB_BUILD_OPTIONS", "") + " nocheck"

    logging.info('--- Running "{}"'.format(" ".join(sbuild_cmd)))
    sbuild_out = subprocess.DEVNULL  # type: Optional[int]
    if sbuild_output_to_terminal:
        sbuild_out = sys.stderr.fileno()

    try:
        subprocess.run(
            sbuild_cmd,
            env=env,
            stdout=sbuild_out,
            stderr=sbuild_out,
            check=True,
        )
    except subprocess.CalledProcessError as e:
        logging.error("--- Error running sbuild")
        pattern = os.path.join(
            SBUILD_DIR, "ubuntu-advantage-tools_*_amd64.build"
        )
        sbuild_log = [f for f in glob.glob(pattern)]

        if sbuild_log:
            with open(sbuild_log[0], "r") as sbuild_log_file:
                logging.error("%s", sbuild_log_file.read())

        raise e

    logging.info("--- Successfully ran sbuild")

    for f in os.listdir(SBUILD_DIR):
        if f.endswith(".deb"):
            if "ubuntu-pro-client-l10n" in f:
                dest = deb_paths.ubuntu_pro_client_l10n
            elif "ubuntu-pro-client" in f:
                dest = deb_paths.ubuntu_pro_client
            elif "ubuntu-pro-auto-attach" in f:
                dest = deb_paths.ubuntu_pro_image_auto_attach
            elif "ubuntu-advantage-tools" in f:
                dest = deb_paths.ubuntu_advantage_tools
            elif "ubuntu-advantage-pro" in f:
                dest = deb_paths.ubuntu_advantage_pro
            else:
                continue
            shutil.copy(os.path.join(SBUILD_DIR, f), dest)

    return deb_paths


class SafeLoaderWithoutDatetime(yaml.SafeLoader):
    yaml_implicit_resolvers = {
        k: [r for r in v if r[0] != "tag:yaml.org,2002:timestamp"]
        for k, v in yaml.SafeLoader.yaml_implicit_resolvers.items()
    }


def _replace_and_log(s, old, new, logger_fn):
    logger_fn('replacing "{}" with "{}"'.format(old, new))
    return s.replace(old, new)


def process_template_vars(
    context, template: str, logger_fn: Optional[Callable] = None, shown=False
) -> str:
    if logger_fn is None:
        logger_fn = logging.info

    processed_template = template

    for match in re.finditer(r"\$behave_var{([\w\s\-\+]*)}", template):
        args = match.group(1).split()
        function_name = args[0]
        if function_name == "version":
            if context.pro_config.check_version:
                processed_template = _replace_and_log(
                    processed_template,
                    match.group(0),
                    context.pro_config.check_version,
                    logger_fn,
                )
        elif function_name == "machine-ip":
            if args[1] in context.machines:
                processed_template = _replace_and_log(
                    processed_template,
                    match.group(0),
                    context.machines[args[1]].instance.ip,
                    logger_fn,
                )
        elif function_name == "machine-name":
            if args[1] in context.machines:
                processed_template = _replace_and_log(
                    processed_template,
                    match.group(0),
                    context.machines[args[1]].instance.name,
                    logger_fn,
                )
        elif function_name == "cloud":
            if args[1] in context.machines:
                processed_template = _replace_and_log(
                    processed_template,
                    match.group(0),
                    context.machines[args[1]].cloud,
                    logger_fn,
                )
        elif function_name == "today":
            dt = datetime.datetime.utcnow()
            if len(args) == 2:
                offset = int(args[1])
                dt = dt + datetime.timedelta(days=offset)
            dt_str = dt.strftime("%Y-%m-%dT00:00:00Z")
            processed_template = _replace_and_log(
                processed_template,
                match.group(0),
                dt_str,
                logger_fn,
            )
        elif function_name == "contract_token_staging":
            processed_template = _replace_and_log(
                processed_template,
                match.group(0),
                context.pro_config.contract_token_staging,
                logger_fn,
            )
        elif function_name == "contract_token":
            processed_template = _replace_and_log(
                processed_template,
                match.group(0),
                context.pro_config.contract_token,
                logger_fn,
            )
        elif function_name == "config":
            item = args[1]
            if not shown or item not in context.pro_config.redact_options:
                processed_template = _replace_and_log(
                    processed_template,
                    match.group(0),
                    getattr(context.pro_config, item),
                    logger_fn,
                )
        elif function_name == "stored_var":
            if context.stored_vars.get(args[1]):
                processed_template = _replace_and_log(
                    processed_template,
                    match.group(0),
                    context.stored_vars.get(args[1]),
                    logger_fn,
                )

    return processed_template


def _landscape_api_request(access_key, secret_key, action, action_params):
    timestamp = time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())
    params = {
        "action": action,
        "access_key_id": access_key,
        "signature_method": "HmacSHA256",
        "signature_version": "2",
        "timestamp": timestamp,
        "version": "2011-08-01",
        **action_params,
    }
    method = "GET"
    host = "landscape.canonical.com"
    path = "/api/"

    formatted_params = "&".join(
        quote(k, safe="~") + "=" + quote(v, safe="~")
        for k, v in sorted(params.items())
    )

    to_sign = "{method}\n{host}\n{path}\n{formatted_params}".format(
        method=method,
        host=host,
        path=path,
        formatted_params=formatted_params,
    )
    digest = hmac.new(
        secret_key.encode(), to_sign.encode(), hashlib.sha256
    ).digest()
    signature = b64encode(digest)
    formatted_params += "&signature=" + quote(signature)

    uri = "https://{host}{path}?{params}".format(
        host=host, path=path, params=formatted_params
    )
    request = Request(
        uri,
        headers={"Host": host},
        method=method,
    )
    response = urlopen(request)

    return response.code, json.load(response)


def landscape_reject_all_pending_computers(context):
    access_key = context.pro_config.landscape_api_access_key
    secret_key = context.pro_config.landscape_api_secret_key
    code, pending_computers = _landscape_api_request(
        access_key, secret_key, "GetPendingComputers", {}
    )
    assert code == 200
    if len(pending_computers) > 0:
        reject_params = {
            "computer_ids.{}".format(i + 1): str(computer["id"])
            for i, computer in enumerate(pending_computers)
        }
        code, _resp = _landscape_api_request(
            access_key, secret_key, "RejectPendingComputers", reject_params
        )
        assert code == 200



import argparse
import sys

from uaclient import (
    actions,
    contract,
    event_logger,
    exceptions,
    messages,
    secret_manager,
)
from uaclient.api.u.pro.attach.magic.initiate.v1 import _initiate
from uaclient.api.u.pro.attach.magic.revoke.v1 import (
    MagicAttachRevokeOptions,
    _revoke,
)
from uaclient.api.u.pro.attach.magic.wait.v1 import (
    MagicAttachWaitOptions,
    _wait,
)
from uaclient.cli import cli_util
from uaclient.cli.commands import ProArgument, ProArgumentGroup, ProCommand
from uaclient.cli.parser import HelpCategory
from uaclient.data_types import AttachActionsConfigFile, IncorrectTypeError
from uaclient.entitlements import (
    create_enable_entitlements_not_found_error,
    get_valid_entitlement_names,
)
from uaclient.entitlements.entitlement_status import CanEnableFailure
from uaclient.yaml import safe_load

event = event_logger.get_event_logger()


def _magic_attach(args, *, cfg, **kwargs):
    if args.format == "json":
        raise exceptions.MagicAttachInvalidParam(
            param="--format",
            value=args.format,
        )

    event.info(messages.CLI_MAGIC_ATTACH_INIT)
    initiate_resp = _initiate(cfg=cfg)
    event.info(
        "\n"
        + messages.CLI_MAGIC_ATTACH_SIGN_IN.format(
            user_code=initiate_resp.user_code
        )
    )

    wait_options = MagicAttachWaitOptions(magic_token=initiate_resp.token)

    try:
        wait_resp = _wait(options=wait_options, cfg=cfg)
    except exceptions.MagicAttachTokenError as e:
        event.info(messages.CLI_MAGIC_ATTACH_FAILED)

        revoke_options = MagicAttachRevokeOptions(
            magic_token=initiate_resp.token
        )
        _revoke(options=revoke_options, cfg=cfg)
        raise e

    event.info("\n" + messages.CLI_MAGIC_ATTACH_PROCESSING)
    return wait_resp.contract_token


@cli_util.assert_not_attached
@cli_util.assert_root
@cli_util.assert_lock_file("pro attach")
def action_attach(args, *, cfg, **kwargs):
    if args.token and args.attach_config:
        raise exceptions.CLIAttachTokenArgXORConfig()
    elif not args.token and not args.attach_config:
        token = _magic_attach(args, cfg=cfg)
        enable_services_override = None
    elif args.token:
        token = args.token
        secret_manager.secrets.add_secret(token)
        enable_services_override = None
    else:
        try:
            attach_config = AttachActionsConfigFile.from_dict(
                safe_load(args.attach_config)
            )
        except IncorrectTypeError as e:
            raise exceptions.AttachInvalidConfigFileError(
                config_name=args.attach_config.name, error=e.msg
            )

        token = attach_config.token
        enable_services_override = attach_config.enable_services

    allow_enable = args.auto_enable and enable_services_override is None

    try:
        actions.attach_with_token(cfg, token=token, allow_enable=allow_enable)
    except exceptions.ConnectivityError:
        raise exceptions.AttachError()
    else:
        ret = 0
        if enable_services_override is not None and args.auto_enable:
            found, not_found = get_valid_entitlement_names(
                enable_services_override, cfg
            )
            for name in found:
                ent_ret, reason = actions.enable_entitlement_by_name(cfg, name)
                if not ent_ret:
                    ret = 1
                    if (
                        reason is not None
                        and isinstance(reason, CanEnableFailure)
                        and reason.message is not None
                    ):
                        event.info(reason.message.msg)
                        event.error(
                            error_msg=reason.message.msg,
                            error_code=reason.message.name,
                            service=name,
                        )
                else:
                    event.service_processed(name)

            if not_found:
                error = create_enable_entitlements_not_found_error(
                    not_found, cfg=cfg
                )
                event.info(error.msg, file_type=sys.stderr)
                event.error(error_msg=error.msg, error_code=error.msg_code)
                ret = 1

        contract_client = contract.UAContractClient(cfg)
        contract_client.update_activity_token()

        cli_util.post_cli_attach(cfg)
        return ret


attach_command = ProCommand(
    "attach",
    help=messages.CLI_ROOT_ATTACH,
    description=messages.CLI_ATTACH_DESC,
    action=action_attach,
    preserve_description=True,
    help_category=HelpCategory.QUICKSTART,
    help_position=2,
    argument_groups=[
        ProArgumentGroup(
            arguments=[
                ProArgument(
                    "token", help=messages.CLI_ATTACH_TOKEN, nargs="?"
                ),
                ProArgument(
                    "--no-auto-enable",
                    help=messages.CLI_ATTACH_NO_AUTO_ENABLE,
                    action="store_false",
                    dest="auto_enable",
                ),
                ProArgument(
                    "--attach-config",
                    help=messages.CLI_ATTACH_ATTACH_CONFIG,
                    type=argparse.FileType("r"),
                ),
                ProArgument(
                    "--format",
                    help=messages.CLI_FORMAT_DESC.format(default="cli"),
                    action="store",
                    choices=["cli", "json"],
                    default="cli",
                ),
            ]
        )
    ],
)
